#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-27
# @Filename: calibration.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum
import os
import pathlib
import time

from typing import TYPE_CHECKING, Any, Literal, Self, cast

from astropy import time as ap_time
from pydantic import BaseModel, Field, model_validator

from sdsstools import read_yaml_file

from gort.exceptions import OverwatcherError
from gort.overwatcher.core import OverwatcherModule, OverwatcherModuleTask
from gort.recipes.calibrations import ExposureTimeTooLong
from gort.tools import (
    add_night_log_comment,
    cancel_task,
    decap,
    redis_client_sync,
    try_or_pass,
)


if TYPE_CHECKING:
    from gort.overwatcher.ephemeris import EphemerisModel
    from gort.overwatcher.helpers.notifier import NotificationLevel
    from gort.overwatcher.overwatcher import Overwatcher


__all__ = ["CalibrationsOverwatcher"]


ETC_DIR = pathlib.Path(__file__).parent / "../etc/"

PathType = os.PathLike | str | pathlib.Path
TimeModeType = Literal["secs_after_sunset", "secs_before_sunrise", "jd", "utc"]
CalsDataType = PathType | list["CalibrationModel"] | list[dict[str, Any]]


class CalibrationModel(BaseModel):
    """Calibration model."""

    name: str = Field(
        ...,
        title="The name of the calibration.",
    )
    recipe: str = Field(
        ...,
        title="The recipe to use.",
    )
    min_start_time: float | None = Field(
        default=None,
        title="The minimum start time. The format depends on time_mode.",
    )
    max_start_time: float | None = Field(
        default=None,
        title="The maximum start time. The format depends on time_mode.",
    )
    time_mode: TimeModeType | None = Field(
        default=None,
        title="The time mode for the calibration.",
    )
    after: str | None = Field(
        default=None,
        title="Run after this calibration.",
    )
    required: bool = Field(
        default=True,
        title="Whether the calibration is required. Currently not used.",
    )
    dome: Literal["open", "closed"] | None = Field(
        default=None,
        title="Whether the dome should be open during the calibration. A null value "
        "will keep the dome in the current position.",
    )
    close_dome_after: bool = Field(
        default=False,
        title="Whether the dome should be closed after the calibration.",
    )
    abort_observing: bool = Field(
        default=False,
        title="Whether observing should be immediately aborted to "
        "allow the calibration to start.",
    )
    priority: int = Field(
        default=5,
        title="The priority of the calibration. Currently not used.",
    )
    max_try_time: float = Field(
        default=300,
        title="The maximum time in seconds to attempt the calibration if it fails. "
        "If max_start_time is reached during this period, the calibrations fails.",
    )
    allow_post_observing_recovery: bool = Field(
        default=True,
        title="Whether the calibration can be run after observing has finished "
        "if it initially failed.",
    )
    disabled: bool = Field(
        default=False,
        title="Whether the calibration is disabled.",
    )

    @model_validator(mode="after")
    def validate_start_time(self) -> Self:
        """Ensure that ``after`` or ``min_start_time`` are defined."""

        if not self.after and self.min_start_time is None:
            raise OverwatcherError("min_start_time or after are required.")

        return self


class CalibrationState(enum.StrEnum):
    """The state of a calibration."""

    WAITING = "waiting"
    RUNNING = "running"
    RETRYING = "retrying"
    FAILED = "failed"
    CANCELLED = "cancelled"
    DONE = "done"


class Calibration:
    """Keeps track of the state of a calibration."""

    def __init__(
        self,
        schedule: CalibrationSchedule,
        calibration: CalibrationModel | dict[str, Any],
    ):
        self.schedule = schedule
        self.ephemeris = self.schedule.ephemeris

        self.model = (
            calibration
            if isinstance(calibration, CalibrationModel)
            else CalibrationModel(**calibration)
        )

        self.name = self.model.name
        self.start_time, self.max_start_time = self.get_start_time()

        self.state = CalibrationState.WAITING
        self._start_time_actual: float | None = None

    def get_start_time(self) -> tuple[float | None, float | None]:
        """Determines the actual time at which the calibration will start.

        Also returns the latest time at which the calibration can start (which
        could be ``None`` if there is no limit). All times are returned as UNIX
        timestamps.

        """

        cal = self.model

        if cal.min_start_time is None and cal.max_start_time is None:
            return None, None

        sunrise = self.ephemeris.sunrise
        sunset = self.ephemeris.sunset

        min_start_time: float | None = None
        max_start_time: float | None = None

        # Convert everything to JD for now.
        if cal.time_mode == "secs_after_sunset":
            if cal.min_start_time is not None:
                min_start_time = sunset + cal.min_start_time / 86400
            if cal.max_start_time is not None:
                max_start_time = sunset + cal.max_start_time / 86400

        elif cal.time_mode == "secs_before_sunrise":
            if cal.min_start_time is not None:
                min_start_time = sunrise - cal.min_start_time / 86400
            if cal.max_start_time is not None:
                max_start_time = sunrise - cal.max_start_time / 86400

        elif cal.time_mode == "jd":
            if cal.min_start_time is not None:
                min_start_time = cal.min_start_time
            if cal.max_start_time is not None:
                max_start_time = cal.max_start_time

        elif cal.time_mode == "utc":
            # SJD at the time of observation is always the following midnight.
            # In this case we want to refer times from the previous midnight.
            base_date = ap_time.Time(self.ephemeris.SJD - 1, format="mjd")

            if cal.min_start_time is not None:
                min_time_d = ap_time.TimeDelta(cal.min_start_time * 3600, format="sec")
                min_start_time = (base_date + min_time_d).jd

            if cal.max_start_time is not None:
                max_start_time_c = cal.max_start_time
                if cal.min_start_time and max_start_time_c < cal.min_start_time:
                    max_start_time_c += 24
                max_time_d = ap_time.TimeDelta(max_start_time_c * 3600, format="sec")
                max_start_time = (base_date + max_time_d).jd

        # Now convert JD to UNIX ap_time.
        min_start_time_unix: float | None = None
        max_start_time_unix: float | None = None

        if min_start_time is not None:
            min_start_time_unix = ap_time.Time(min_start_time, format="jd").unix

        if max_start_time is not None:
            max_start_time_unix = ap_time.Time(max_start_time, format="jd").unix

        return min_start_time_unix, max_start_time_unix

    def to_dict(self) -> dict[str, Any]:
        """Returns the calibration as a dictionary."""

        return {
            "name": self.name,
            "state": self.state.name.lower(),
        }

    def is_finished(self):
        """Returns ``True`` if the calibration is done or has failed."""

        if self.state in (
            CalibrationState.DONE,
            CalibrationState.FAILED,
            CalibrationState.CANCELLED,
        ):
            return True

    async def record_state(
        self,
        state: CalibrationState | None = None,
        fail_reason: str = "unespecified reason",
        add_to_night_log: bool = True,
    ):
        """Records the state of the calibration in Redis."""

        if state is not None:
            self.state = state

        if state == CalibrationState.RUNNING:
            self._start_time_actual = time.time()

        with redis_client_sync() as redis:
            key = f"overwatcher:calibrations:{self.schedule.sjd}"

            # Check that the key exists. If it does not, it means that update_schedule()
            # has not been called yet.
            if redis.json().get(key) is None:
                cals_file = self.schedule.cals_overwatcher.cals_file
                await self.schedule.cals_overwatcher.reset(cals_file)

            redis.json().set(key, f".{self.name}.state", self.state.name.lower())

        # Add a comment to the night log section for issues indicating the calibration
        # has failed and the reason for it.
        if state == CalibrationState.FAILED and add_to_night_log:
            if not fail_reason.endswith("."):
                fail_reason += "."

            await add_night_log_comment(
                f"Calibration {self.name} failed. Reason: {decap(fail_reason)}"
                " See log for more details.",
                category="overwatcher",
            )


class CalibrationSchedule:
    """A class to handle how to schedule calibrations.

    Parameters
    ----------
    cals_overwatcher
        The calibrations overwatcher instance that will monitor the schedule.
    cals_data
        The calibration data to use to generate the schedule. Either a YAML file
        with the calibrations to be taken, or a list of ``CalibrationModel`` instances.
    update
        Update the schedule on init.

    """

    def __init__(
        self,
        cals_overwatcher: CalibrationsOverwatcher,
        cals_data: CalsDataType,
        update: bool = True,
    ):
        self.cals_overwatcher = cals_overwatcher
        self.log = self.cals_overwatcher.log

        self.calibrations: list[Calibration] = []

        # These are set by .refresh()
        self.sjd: int = 0
        self.ephemeris: EphemerisModel

        if update:
            self.update_schedule(cals_data=cals_data)

    def get_calibration(self, name: str):
        """Returns a calibration by name."""

        for cal in self.calibrations:
            if cal.name == name:
                return cal

        return None

    def update_schedule(
        self,
        cals_data: CalsDataType | None = None,
        reset: bool = False,
    ):
        """Generates a schedule of calibrations for the night."""

        if not self.cals_overwatcher.overwatcher.ephemeris.ephemeris:
            self.log.error("Ephemeris are not available. Not updating schedule.")
            return

        self.ephemeris = self.cals_overwatcher.overwatcher.ephemeris.ephemeris
        self.sjd = self.cals_overwatcher.overwatcher.ephemeris.sjd

        # Update the calibration models.
        if cals_data is not None:
            if isinstance(cals_data, (list, tuple)):
                cals_data = [
                    CalibrationModel(**cal) if isinstance(cal, dict) else cal
                    for cal in cals_data
                ]
            else:
                cals_data = self.read_file(cals_data)

            self.calibrations = [Calibration(self, cal) for cal in cals_data]

        else:
            pass

        if len(self.calibrations) == 0:
            self.log.warning("No calibrations found.")
            return

        self.log.info(f"Updating calibrations schedule for SJD {self.sjd}.")

        # Get the state of the calibrations from Redis. This is important if for
        # example the overwatcher is restarted after some calibrations have already
        # been taken. We use sync Redis because this should be fast and allows
        # update_schedule() to be called by __init__().
        with redis_client_sync() as redis:
            # The key we use to store the information for this SJD.
            key = f"overwatcher:calibrations:{self.sjd}"

            # Check if the key exists. If it does not, it means we haven't done
            # anything yet for this SJD.
            if not redis.exists(key):
                # Purge all existing data.
                data = {cal.name: cal.to_dict() for cal in self.calibrations}
                redis.json().set(key, "$", data)
            else:
                # Get calibration data from Redis and update the object.
                redis_cal_keys = redis.json().objkeys(key, "$")[0]

                for cal in self.calibrations:
                    if not reset and redis_cal_keys and cal.name in redis_cal_keys:
                        state = cast(str, redis.json().get(key, f".{cal.name}.state"))
                        cal.state = CalibrationState(state.lower())
                    else:
                        redis.json().set(key, f".{cal.name}", cal.to_dict())

    def read_file(self, filename: PathType) -> list[CalibrationModel]:
        """Loads calibrations from a file."""

        cal_data: list = read_yaml_file(filename, return_class=list)  # type: ignore
        if not isinstance(cal_data, (list, tuple)):
            raise ValueError("Calibrations file is badly formatted.")

        return [CalibrationModel(**cal) for cal in cal_data]

    async def get_next(self):
        """Returns the next calibration or ``None`` if no calibration is due.

        Takes into account open dome buffer time.

        """

        # First check if any calibration is running. If that's the case,
        # return it immediately.
        for cal in self.calibrations:
            if cal.state == CalibrationState.RUNNING:
                return cal

        overwatcher = self.cals_overwatcher.overwatcher
        open_dome_buffer = overwatcher.config["overwatcher.schedule.open_dome_buffer"]

        # Do not allow calibrations if the Overwatcher is troubleshooting.
        if overwatcher.state.troubleshooting:
            return None

        now: float = time.time()
        done_cals: set[str] = set()

        for cal in sorted(
            self.calibrations,
            key=lambda x: x.model.priority,
            reverse=True,
        ):
            if cal.name in done_cals or cal.is_finished():
                done_cals.add(cal.name)
                continue

            if cal.state == CalibrationState.RETRYING:
                return cal

            if cal.model.disabled:
                continue

            # If it's too late to start the calibration, skip it.
            if cal.max_start_time is not None and now > cal.max_start_time:
                await overwatcher.notify(
                    f"Skipping calibration {cal.name} as it's too late to start it.",
                    level="error",
                )
                await cal.record_state(
                    CalibrationState.FAILED,
                    fail_reason="too late to start the calibration.",
                )
                continue

            if cal.model.after is not None:
                if cal.model.after not in done_cals:
                    continue

            if cal.start_time is None:
                if cal.model.after is not None:
                    return cal

                self.log.warning(
                    f"Calibration {cal.name} has no start time and after is null. "
                    "This should not happen."
                )
                await cal.record_state(CalibrationState.FAILED, add_to_night_log=False)
                continue

            # If the calibration start time is too fast in the future even considering
            # that we may need to move the dome, continue. This is mostly to avoid
            # too many calls to gort.enclosure.is_open() which are kind of slow.
            if now < cal.start_time - open_dome_buffer:
                continue

            # If the calibration requires a specific position for the dome,
            # check the current dome position and adjust the start time if needed.
            min_start_time_dome: float = cal.start_time
            if (dome_requested := cal.model.dome) is not None:
                dome_open = await overwatcher.dome.is_opening()
                dome_requested_bool = True if dome_requested == "open" else False

                if dome_requested_bool is not dome_open:
                    min_start_time_dome = cal.start_time - open_dome_buffer

            if now >= min_start_time_dome:
                return cal

        return None


class CalibrationsMonitor(OverwatcherModuleTask["CalibrationsOverwatcher"]):
    """Monitors the calibrations schedule."""

    name = "calibrations_monitor"
    keep_alive = True
    restart_on_error = True

    async def task(self):
        """Runs the calibration monitor."""

        notify = self.overwatcher.notify

        while True:
            await asyncio.sleep(10)

            next_calibration = await self.module.schedule.get_next()
            allow_calibrations = self.module.overwatcher.state.allow_calibrations

            if next_calibration is not None and allow_calibrations:
                name = next_calibration.name
                try:
                    self.module._calibration_task = asyncio.create_task(
                        self.module.run_calibration(next_calibration)
                    )
                    await self.module._calibration_task

                except asyncio.CancelledError:
                    if not next_calibration.is_finished():
                        await notify(
                            f"Calibration {name} has been cancelled.",
                            level="warning",
                        )
                        await next_calibration.record_state(CalibrationState.CANCELLED)

                except ExposureTimeTooLong as ee:
                    # This is a particular case when the exposure time for a twilight
                    # flat is too long. This is not exactly an error, since we now it
                    # often happens and that point we should have exposed most fibres,
                    # so we just mark the calibration as done and issue a notification,
                    # but not an error.
                    if not next_calibration.is_finished():
                        await notify(str(ee.raw_message), level="warning")
                        await notify(f"Calibration {name} is done.")

                except Exception as ee:
                    if not next_calibration.is_finished():
                        await notify(
                            f"Error running calibration {name}: {decap(ee)}",
                            level="error",
                        )
                        await next_calibration.record_state(
                            CalibrationState.FAILED,
                            fail_reason=str(ee),
                        )

                finally:
                    close_dome_after = next_calibration.model.close_dome_after
                    dome_closed = await self.module.overwatcher.dome.is_closing()
                    dome_locked = self.overwatcher.dome.locked

                    if next_calibration.state == CalibrationState.RETRYING:
                        # If the calibration is retrying, we do not close the dome
                        # or mark it as done. Move to the next iteration of the loop.
                        continue

                    if close_dome_after and not dome_closed:
                        if not dome_locked:
                            await notify(f"Closing the dome after calibration {name}.")
                            await try_or_pass(self.overwatcher.dome.close())
                        else:
                            await notify(
                                "Dome is locked. Not closing it after calibration.",
                                level="warning",
                            )

                    if not next_calibration.is_finished():
                        await next_calibration.record_state(CalibrationState.DONE)


class CalibrationsOverwatcher(OverwatcherModule):
    """Calibrations overwatcher module."""

    name = "calibration"
    delay = 3

    tasks = [CalibrationsMonitor()]

    def __init__(
        self,
        overwatcher: Overwatcher,
        cals_file: str | pathlib.Path | None = None,
    ):
        super().__init__(overwatcher)

        self._default_cals_file = ETC_DIR / "calibrations.yaml"
        self.cals_file = pathlib.Path(cals_file or self._default_cals_file)

        self.schedule = CalibrationSchedule(self, self.cals_file, update=False)

        self._calibration_task: asyncio.Task | None = None

    async def reset(self, cals_file: str | pathlib.Path | None = None):
        """Resets the list of calibrations for a new SJD.

        This method is usually called by the ephemeris overwatcher when a new SJD
        is detected.

        """

        while self.overwatcher.ephemeris.ephemeris is None:
            self.log.error("Ephemeris is not available. Cannot reset calibrations.")
            await asyncio.sleep(5)

        if cals_file is not None:
            self.cals_file = cals_file

        try:
            self.schedule.update_schedule(self.cals_file)
        except Exception as ee:
            self.log.error(f"Error updating calibrations schedule: {decap(ee)}")

    def get_running_calibration(self):
        """Returns the calibration currently running."""

        for cal in self.schedule.calibrations:
            if cal.state == CalibrationState.RUNNING:
                return cal

    def is_calibrating(self):
        """Returns True if a calibration is currently running."""

        return self.get_running_calibration() is not None

    async def run_calibration(self, calibration: Calibration):
        """Runs a calibration."""

        notify = self.overwatcher.notify

        name = calibration.name
        max_start_time = calibration.max_start_time
        dome = calibration.model.dome
        recipe = calibration.model.recipe

        if calibration.is_finished():
            self.log.warning(f"Calibration {name} is already finished. Skipping.")
            return

        if self.overwatcher.state.dry_run:
            self.log.warning(f"Dry-run mode. Marking calibration {name} done.")
            await calibration.record_state(CalibrationState.DONE)
            return

        running_cal = self.get_running_calibration()
        if running_cal is not None:
            await self._fail_calibration(
                calibration,
                f"Cannot run {name}. A calibration is already running!",
                level="warning",
            )
            return

        if calibration.state == CalibrationState.RUNNING:
            self.log.warning(
                f"Calibration {name} is recorded as running but it is not. "
                "Setting its status to WAITING."
            )
            await calibration.record_state(CalibrationState.WAITING)

        if not self.overwatcher.state.allow_calibrations:
            await self._fail_calibration(
                calibration,
                f"Cannot run {name}. Calibrations are disabled.",
                level="error",
            )
            return

        if calibration.state != CalibrationState.RETRYING:
            await notify(f"Running calibration {name}.")

        await calibration.record_state(CalibrationState.RUNNING)

        now = time.time()
        if max_start_time is not None and now > max_start_time:
            await notify(
                f"Skipping calibration {name} as it's too late to start it.",
                level="warning",
            )
            await calibration.record_state(
                CalibrationState.FAILED,
                fail_reason="too late to start the calibration.",
            )
            return

        if self.overwatcher.state.observing:
            immediate = calibration.model.abort_observing

            await notify(f"Aborting observations to run calibration {name}.")
            await self.overwatcher.observer.stop_observing(
                reason="Scheduled calibration",
                immediate=immediate,
                block=True,
            )

        if dome is not None:
            dome_new = True if dome == "open" else False
            dome_current = await self.overwatcher.dome.is_opening()
            needs_dome_change = dome_new != dome_current

            if needs_dome_change:
                if await self.overwatcher.dome.is_moving():
                    try:
                        await notify("The dome is moving. Waiting for it to stop.")
                        await asyncio.wait_for(
                            self.overwatcher.dome.wait_until_idle(),
                            timeout=180,
                        )
                    except asyncio.TimeoutError:
                        await self._fail_calibration(
                            calibration,
                            f"Cannot move dome for calibration {name}. "
                            "Dome is still moving after 3 minutes.",
                            level="error",
                            fail_now=True,
                        )
                        return

                if not self.overwatcher.state.safe:
                    # Fail immediately if the weather is not safe. There is little
                    # chance that conditions will improve in the next few minutes.
                    await self._fail_calibration(
                        calibration,
                        f"Cannot move dome for {name}. Weather is not safe.",
                        level="error",
                        fail_now=True,
                    )
                    return

                if not self.overwatcher.state.enabled:
                    await self._fail_calibration(
                        calibration,
                        f"Cannot move dome for {name}. Overwatcher is disabled.",
                        level="error",
                    )
                    return

                if dome == "open" and not dome_current:
                    await notify(f"Opening the dome for calibration {name}.")
                    await self.overwatcher.dome.open()
                elif dome == "closed" and dome_current:
                    await notify(f"Closing the dome for calibration {name}.")
                    await self.overwatcher.dome.close()

        await notify(f"Running recipe {recipe!r} for calibration {name}.")
        await self.overwatcher.gort.execute_recipe(recipe)

        await notify(f"Calibration {name} is done.")
        # We do not set the state to DONE here. That is done in the monitor task
        # after closing the dome. That prevents collisions with the main task that
        # is monitoring for daytime conditions.

    async def cancel(self):
        """Cancel the running calibration."""

        notify = self.overwatcher.notify

        running_calibration = self.get_running_calibration()

        if (
            self._calibration_task is not None
            and not self._calibration_task.done()
            and running_calibration
        ):
            name = running_calibration.name

            await notify(f"Cancelling calibration {name}.", level="warning")
            self._calibration_task = await cancel_task(self._calibration_task)

            await running_calibration.record_state(CalibrationState.CANCELLED)

            # Ensure we close the dome. This is allowed even
            # if the overwatcher is disabled.
            if running_calibration.model.close_dome_after:
                try:
                    await notify(f"Closing the dome after calibration {name}.")
                    await self.overwatcher.dome.close()
                except Exception as ee:
                    await notify(
                        f"Error closing the dome after calibration {name}: {decap(ee)}",
                        level="error",
                    )

    async def _fail_calibration(
        self,
        calibration: Calibration,
        message: str,
        level: NotificationLevel = "error",
        repeat_notifications: bool = False,
        fail_now: bool = False,
    ):
        """Decides whether to fail a calibration or continue trying."""

        name = calibration.name
        max_try_time = calibration.model.max_try_time

        was_already_retrying = calibration.state == CalibrationState.RETRYING
        start_time = calibration._start_time_actual

        if max_try_time <= 0 or fail_now:
            await self.overwatcher.notify(message, level=level)
            await calibration.record_state(CalibrationState.FAILED, fail_reason=message)
            return

        await calibration.record_state(CalibrationState.RETRYING)

        if not was_already_retrying:
            await self.overwatcher.notify(message, level=level)
            return

        if start_time and time.time() - start_time > max_try_time:
            await self.overwatcher.notify(
                f"Maximum try time reached for calibration {name}. "
                "Failing the calibration now.",
                level=level,
            )
            await calibration.record_state(
                CalibrationState.FAILED,
                fail_reason="maximum time reached trying to run the calibration. "
                f"Original error: {decap(message)}",
            )
            return

        if repeat_notifications:
            await self.overwatcher.notify(message, level=level)

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-08
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


import asyncio
import os

import click

from sdsstools.daemonizer import DaemonGroup, cli_coro

from gort.tools import get_gort_client


@click.group()
@click.option(
    "--override-overwatcher",
    is_flag=True,
    help="Performs the operation even if the Overwatcher is running.",
)
def gort(override_overwatcher: bool = False):
    """GORT CLI."""

    os.environ["GORT_OVERRIDE_OVERWATCHER"] = "1" if override_overwatcher else "0"


@gort.group(cls=DaemonGroup, prog="gort-overwatcher-actor", workdir=os.getcwd())
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    help="Path to the configuration file. Must include a "
    "section called 'actor' or 'overwatcher.actor'.",
)
@click.option(
    "--verbosity",
    "-v",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        case_sensitive=False,
    ),
    help="Verbosity level for console output.",
)
@click.option("--dry-run", is_flag=True, help="Runs the actor in dry-run mode.")
@cli_coro()
async def overwatcher(
    config: str | None = None,
    dry_run: bool = False,
    verbosity: str | None = None,
):
    """Starts the overwatcher."""

    from gort import config as gort_config
    from gort.overwatcher.actor import OverwatcherActor

    if config is not None:
        gort_config.load(config)

    actor_config = gort_config["overwatcher.actor"]
    if verbosity:
        actor_config["console_verbosity"] = verbosity

    actor = OverwatcherActor.from_config(actor_config, dry_run=dry_run)
    await actor.start()

    while True:
        await asyncio.sleep(5)
        continue


@gort.command()
@click.option(
    "--open-enclosure/--no-open-enclosure",
    is_flag=True,
    default=True,
    help="Open the enclosure.",
)
@click.option(
    "--confirm-open/--no-confirm-open",
    is_flag=True,
    default=True,
    help="Request confirmation before opening the enclosure.",
)
@click.option(
    "--focus/--no-focus",
    is_flag=True,
    default=True,
    help="Focus the telescopes.",
)
@cli_coro()
async def startup(
    open_enclosure: bool = True,
    confirm_open: bool = True,
    focus: bool = True,
):
    """Runs the startup sequence."""

    async with get_gort_client() as gort:
        await gort.startup(
            open_enclosure=open_enclosure,
            confirm_open=confirm_open,
            focus=focus,
        )


@gort.command()
@click.option(
    "--park/--no-park",
    is_flag=True,
    default=True,
    help="Parks the telescopes.",
)
@cli_coro()
async def shutdown(park: bool = True):
    """Runs the shutdown sequence."""

    async with get_gort_client() as gort:
        await gort.shutdown(park_telescopes=park)


@gort.command()
@click.option(
    "--readout/--no-readout",
    is_flag=True,
    default=False,
    help="Read the spectrographs if an exposure is pending.",
)
@cli_coro()
async def cleanup(readout: bool = False):
    """Runs the cleanup sequence."""

    async with get_gort_client() as gort:
        await gort.cleanup(readout=readout)


@gort.command()
@click.argument("RECIPE", type=str)
@cli_coro()
async def recipe(recipe: str):
    """Runs a recipe with its default options."""

    async with get_gort_client() as gort:
        await gort.execute_recipe(recipe)


@gort.command()
@cli_coro()
async def open():
    """Opens the dome."""

    async with get_gort_client() as gort:
        await gort.enclosure.open()


@gort.command()
@cli_coro()
async def close():
    """Closes the dome."""

    async with get_gort_client() as gort:
        await gort.enclosure.close(force=True)


@gort.command()
@cli_coro()
async def focus():
    """Focus the telescopes."""

    async with get_gort_client() as gort:
        await gort.guiders.focus()


@gort.command()
@cli_coro()
async def observe():
    """Runs the observe loop."""

    async with get_gort_client() as gort:
        await gort.observe(show_progress=True)


@gort.command(name="pointing-model")
@click.argument(
    "TELESCOPES",
    type=click.Choice(["sci", "spec", "skye", "skyw"]),
    nargs=-1,
    required=False,
)
@click.option(
    "-n",
    "--n-points",
    default=50,
    help="Number of points to sample.",
)
@click.option(
    "-a",
    "--alt-range",
    nargs=2,
    type=float,
    default=(40, 85),
    help="Altitude range.",
)
@click.option(
    "-z",
    "--az-range",
    nargs=2,
    type=float,
    default=(0, 355),
    help="Azimuth range.",
)
@click.option(
    "-f",
    "--filename",
    type=click.Path(exists=False, dir_okay=False),
    help="Path where to save the pointing data.",
)
@click.option(
    "-m",
    "--home",
    is_flag=True,
    help="Home the telescope before acquiring data.",
)
@click.option(
    "-P/-p",
    "--add-points/--no-add-points",
    is_flag=True,
    default=True,
    help="Add points to the PWI model.",
)
@click.option(
    "--only-slew",
    is_flag=True,
    help="Only slews the telescopes but does not take pointing data.",
)
@cli_coro()
async def pointing_model(
    telescopes: tuple[str, ...] | None,
    alt_range: tuple[float, float] = (40, 85),
    az_range: tuple[float, float] = (0, 355),
    filename: str | None = None,
    n_points: int = 50,
    home: bool = False,
    add_points: bool = True,
    only_slew: bool = False,
):
    """Acquires a pointing model."""

    from gort.pointing import pointing_model

    if telescopes is None or len(telescopes) == 0:
        telescopes = ("sci", "spec", "skye", "skyw")

    await pointing_model(
        filename,
        n_points,
        alt_range,
        az_range,
        telescopes=telescopes,
        home=home,
        add_points=add_points,
        calculate_offset=not only_slew,
    )


def main():
    gort(auto_envvar_prefix="GORT")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-12-29
# @Filename: health.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import time

from typing import TYPE_CHECKING, Sequence

from lvmopstools.retrier import Retrier

from gort import config
from gort.tools import (
    get_lvmapi_route,
    kubernetes_list_deployments,
    kubernetes_restart_deployment,
)


if TYPE_CHECKING:
    from gort.gort import Gort


@Retrier(max_attempts=3, delay=1)
async def ping_actors() -> dict[str, bool]:
    """Pings all actors in the system."""

    return await get_lvmapi_route("/actors/ping")


async def get_actor_ping(
    discard_disabled: bool = False,
    discard_overwatcher: bool = False,
) -> dict[str, bool]:
    """Returns a list of failed actors."""

    disabled_actors = config["overwatcher.disabled_actors"] or []

    actor_status = await ping_actors()

    for actor in disabled_actors:  # Ignore actors we know are disabled.
        if discard_disabled and actor in actor_status:
            actor_status.pop(actor)

    if discard_overwatcher and "lvm.overwatcher" in actor_status:
        actor_status.pop("lvm.overwatcher")

    return actor_status


async def restart_actors(
    actors: str | Sequence[str],
    gort: Gort,
    post_restart_actions: bool = True,
    timeout: float = 60,
) -> None:
    """Restarts one or more actors.

    This is a high-level function that tries to restart a series of actors, along
    with their deployments, in an optimal way. If multiple actors correspond to the
    same deployment, only one deployment restart is triggered.

    Parameters
    ----------
    actors
        The actor or actors to restart.
    gort
        The Gort instance. Used to refresh the command set in the affected actors.
    post_restart_actions
        Whether to run post-restart actions. This may include resetting spectrographs,
        homing devices, etc.
    timeout
        The timeout in seconds to wait for the restart to complete.

    """

    if isinstance(actors, str):
        actors = [actors]

    try:
        actor_to_deployment = await get_lvmapi_route("/actors/actor-to-deployment")
        deployment_to_actors = await get_lvmapi_route("/actors/deployment-to-actors")
    except Exception as ee:
        raise RuntimeError("Unable to retrieve actor-deployment mappings.") from ee

    deployments_to_restart = set()
    for actor in actors:
        if actor not in actor_to_deployment:
            raise ValueError(f"Actor {actor} not found in actor-to-deployment mapping.")
        deployments_to_restart.add(actor_to_deployment[actor])

    for deployment in deployments_to_restart:
        if deployment not in deployment_to_actors:
            raise ValueError(f"Actors for deployment {deployment!r} cannot be found.")

        await restart_deployment(
            gort,
            deployment,
            deployment_to_actors[deployment],
            timeout=timeout,
        )

    for actor in actors:
        if actor in gort.actors:
            gort.log.debug(f"Refreshing command list for actor {actor!r} in Gort.")
            await gort.actors[actor].refresh()
        else:
            gort.log.warning(f"Actor {actor!r} not found in Gort. Adding it.")
            gort.add_actor(actor)
            await gort.actors[actor].init()

    if post_restart_actions:
        gort.log.debug("Running post-restart actions.")
        await run_post_restart_actions(gort, actors)


async def restart_deployment(
    gort: Gort,
    deployment: str,
    actors: list[str],
    timeout: float = 60,
):
    """Restarts a deployment, waiting until all its actors ping."""

    start_time = time.time()

    gort.log.warning(f"Restarting deployment {deployment}.")
    await kubernetes_restart_deployment(deployment)

    await asyncio.sleep(5)

    while True:
        if deployment in await kubernetes_list_deployments():
            gort.log.info(f"Deployment {deployment} restarted.")
            break

        if time.time() - start_time > timeout:
            raise RuntimeError(
                f"Timed out waiting for deployment {deployment!r} to restart."
            )

        await asyncio.sleep(1)

    while True:
        actor_pings = {actor: False for actor in actors}

        for actor in actors:
            ping_cmd = await gort.send_command(actor, "ping", time_limit=2)
            if ping_cmd.status.did_succeed:
                actor_pings[actor] = True

        if all(actor_pings.values()):
            gort.log.info(f"All actors in deployment {deployment} are pinging.")
            break

        if time.time() - start_time > timeout:
            raise RuntimeError(
                f"Timed out waiting actors in deployment {deployment!r} to ping."
            )


async def run_post_restart_actions(gort: Gort, actors: Sequence[str]):
    """Runs post-restart actions for a list of actors."""

    # Start with the easy ones. If the AG actors or specs have been restarted,
    # reset them.
    agcam_restarted = any(".agcam" in actor for actor in actors)
    scp_restarted = any("lvmscp" in actor for actor in actors)

    if agcam_restarted:
        gort.log.info("AG actors restarted. Reconnecting AG cameras.")
        await gort.ags.reconnect()

    if scp_restarted:
        gort.log.info("SCP actors restarted. Resetting spectrographs.")
        await gort.specs.abort()
        await gort.specs.reset(full=True)

    # Now re-home telescope devices.
    for telescope in ["sci", "spec", "skye", "skyw"]:
        for device in ["km", "fibsel", "foc"]:
            if f"lvm.{telescope}.{device}" not in actors:
                continue

            gort.log.info(f"Re-homing {device!r} on telescope {telescope}.")

            tel_dev = gort.telescopes[telescope]
            dev = getattr(tel_dev, "focuser" if device == "foc" else device)
            if dev is None:
                continue

            try:
                await dev.home()
            except Exception:
                gort.log.warning(f"Failed to home {device!r} on telescope {telescope}.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-08-03
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib

from clu.actor import AMQPActor
from clu.command import Command

import gort
from gort.overwatcher.overwatcher import Overwatcher

from .commands import overwatcher_cli


__all__ = ["OverwatcherActor", "OverwatcherCommand"]


class OverwatcherActor(AMQPActor):
    """An actor that watches over other actors!"""

    parser = overwatcher_cli
    parser_raise_on_error = True

    def __init__(
        self,
        *args,
        dry_run: bool = False,
        console_verbosity: str = "WARNING",
        **kwargs,
    ):
        gort_root = pathlib.Path(gort.__file__).parent
        schema = gort_root / "etc" / "actor_schema.json"

        super().__init__(*args, schema=schema, version=gort.__version__, **kwargs)

        self.overwatcher = Overwatcher(dry_run=dry_run, verbosity=console_verbosity)

        self.log.info("OverwatcherActor initialised.")

    async def start(self, **kwargs):
        """Starts the overwatcher and actor."""

        await self.overwatcher.run()
        await super().start(**kwargs)

        # Recreate the GORT exception hooks. We want exceptions to be logged
        # to the GORT log and error events to be emitted.
        self.overwatcher.gort._setup_exception_hooks()
        self.overwatcher.gort._setup_async_exception_hooks()


OverwatcherCommand = Command[OverwatcherActor]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-07-09
# @Filename: observer.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import logging
import re
import signal
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial, wraps
from time import time

from typing import TYPE_CHECKING, Any, Callable, Coroutine, TypedDict

from astropy.time import Time

from sdsstools.utils import GatheringTaskGroup

from gort.enums import Event, GuiderStatus, ObserverStageStatus
from gort.exceptions import (
    ErrorCode,
    GortError,
    GortObserverCancelledError,
    GortObserverError,
)
from gort.exposure import Exposure
from gort.tile import Coordinates, Tile
from gort.tools import (
    GuiderMonitor,
    cancel_task,
    decap,
    handle_signals,
    insert_to_database,
    register_observation,
    run_in_executor,
)
from gort.transforms import fibre_slew_coordinates, wrap_pa_hex


if TYPE_CHECKING:
    from gort.gort import Gort


__all__ = ["GortObserver"]


class InterrupHandlerHelper:
    """Helper for handling interrupts"""

    def __init__(self):
        self.observer: GortObserver | None = None
        self._callback: Callable | None = None

    def run_callback(self):
        if self._callback is None:
            return

        if self.observer:
            if (exposure := self.observer._current_exposure) is not None:
                exposure.error = True
                exposure.set_result(exposure)
                exposure.stop_timer()

            self.observer.gort.log.warning("Running cleanup due to keyboard interrupt.")

        self._callback()

    def set_callback(self, cb: Callable | None):
        self._callback = cb


interrupt_helper = InterrupHandlerHelper()
interrupt_signals = [signal.SIGINT, signal.SIGTERM]


class StagesDict(TypedDict):
    """A dictionary of observer stages."""

    slew: ObserverStageStatus
    acquire: ObserverStageStatus
    expose: ObserverStageStatus


def register_stage_status(coro):
    """Records the status of the stage."""

    @wraps(coro)
    async def wrapper(*args, **kwargs):
        self: GortObserver = args[0]
        stage = coro.__name__

        tile = self._tile
        tile_id = tile.tile_id if tile else None
        dither_position = self.dither_position if tile else None

        payload = {
            "tile_id": tile_id,
            "dither_position": dither_position,
            "stage": stage,
        }

        if self.cancelling:
            raise GortObserverCancelledError()

        self.stages[stage] = ObserverStageStatus.RUNNING
        await self.gort.notify_event(Event.OBSERVER_STAGE_RUNNING, payload=payload)

        try:
            result = await coro(*args, **kwargs)
        except Exception:
            self.stages[stage] = ObserverStageStatus.FAILED
            await self.gort.notify_event(Event.OBSERVER_STAGE_FAILED, payload=payload)
            raise

        self.stages[stage] = ObserverStageStatus.DONE
        await self.gort.notify_event(Event.OBSERVER_STAGE_DONE, payload=payload)

        return result

    return wrapper


class GortObserver:
    """A class to handle tile observations.

    Parameters
    ----------
    gort
        The instance of :obj:`.Gort` used to communicate with the devices.
    tile
        The :obj:`.Tile` with the information about the observation.
    mask_positions_pattern
        The ``spec`` fibre mask positions to use.
    on_interrupt
        Callback to be called when the observation is interrupted.

    """

    def __init__(
        self,
        gort: Gort,
        tile: Tile | None = None,
        mask_positions_pattern: str = "P1-*",
        on_interrupt: Callable | None = None,
    ):
        self.gort = gort
        self._tile = tile

        self.mask_positions = self._get_mask_positions(mask_positions_pattern)

        self.guide_task: asyncio.Future | None = None
        self.guider_monitor = GuiderMonitor(self.gort)

        self.standards: Standards | None = None

        self.dither_position: int = 0
        self._current_exposure: Exposure | None = None

        self.overheads: dict[str, dict[str, int | float]] = {}

        self.stages: StagesDict = {
            "slew": ObserverStageStatus.WAITING,
            "acquire": ObserverStageStatus.WAITING,
            "expose": ObserverStageStatus.WAITING,
        }

        self.cancelling: bool = False

        interrupt_helper.set_callback(on_interrupt)
        interrupt_helper.observer = self

        # Necessary because this sets the standards
        # and dither position if the tile is not null.
        self.reset(tile)

    def reset(
        self,
        tile: Tile | None = None,
        on_interrupt: Callable | None = None,
        reset_stages: bool = True,
    ):
        """Resets the observer."""

        self._tile = tile
        self.standards = Standards(self, tile) if tile else None
        self.dither_position = tile.sci_coords.dither_position if tile else 0

        self.guide_task = None
        self.guider_monitor.reset()

        self._current_exposure = None
        self.overheads = {}

        self.cancelling = False

        if reset_stages:
            self.stages: StagesDict = {
                "slew": ObserverStageStatus.WAITING,
                "acquire": ObserverStageStatus.WAITING,
                "expose": ObserverStageStatus.WAITING,
            }

        if on_interrupt is not None:
            interrupt_helper.set_callback(on_interrupt)

    @property
    def tile(self):
        """Returns the current tile being observed."""

        if not self._tile:
            raise GortObserverError("No tile has been set.")

        return self._tile

    def set_tile(self, tile: Tile):
        """Sets the current tile. Implies reset."""

        self.reset(tile)

    def is_running(self):
        """Returns :obj:`True` if the observer is running."""

        return self.get_running_stage() is not None

    def get_running_stage(self):
        """Returns the running stage."""

        for stage, status in self.stages.items():
            if status == ObserverStageStatus.RUNNING:
                return stage

        return None

    def __repr__(self):
        tile_id = self._tile.tile_id if self._tile else "none"
        return f"<GortObserver (tile_id={tile_id})>"

    @property
    def has_standards(self):
        """Returns :obj:`True` if standards will be observed."""

        if not self.standards:
            return False

        return len(self.standards.standards) > 0

    async def observe_tile(
        self,
        tile: Tile | int | None = None,
        ra: float | None = None,
        dec: float | None = None,
        pa: float = 0.0,
        use_scheduler: bool = True,
        dither_position: int | None = None,
        exposure_time: float = 900.0,
        n_exposures: int = 1,
        async_readout: bool = False,
        keep_guiding: bool = False,
        skip_slew_when_acquired: bool = True,
        guide_tolerance: float = 1.0,
        acquisition_timeout: float = 180.0,
        show_progress: bool | None = None,
        run_cleanup: bool = True,
        adjust_focus: bool = True,
        cleanup_on_interrupt: bool = True,
    ) -> tuple[bool, list[Exposure]]:
        """Performs all the operations necessary to observe a tile.

        Parameters
        ----------
        tile
            The ``tile_id`` to observe, or a :obj:`.Tile` object. If not
            provided, observes the next tile suggested by the scheduler
            (requires ``use_scheduler=True``).
        ra,dec
            The RA and Dec where to point the science telescopes. The other
            telescopes are pointed to calibrators that fit the science pointing.
            Cannot be used with ``tile``.
        pa
            Position angle of the IFU. Defaults to PA=0.
        use_scheduler
            Whether to use the scheduler to determine the ``tile_id`` or
            select calibrators.
        dither_position
            The dither position to use. If not provided, uses the tile default
            dither position or zero.
        exposure_time
            The length of the exposure in seconds.
        n_exposures
            Number of exposures to take while guiding.
        async_readout
            Whether to wait for the readout to complete or return as soon
            as the readout begins. If :obj:`False`, the exposure is registered
            but the observation is not finished. This should be :obj:`True`
            during normal science operations to allow the following acquisition
            to occur during readout.
        keep_guiding
            If :obj:`True`, keeps the guider running after the last exposure.
            This should be :obj:`False` during normal science operations.
        skip_slew_when_acquired
            If the tile has been acquired and is guiding, skips the slew,
            modifies the dither position, waits until the new position has
            been acquired, and starts exposing.
        guide_tolerance
            The guide tolerance in arcsec. A telescope will not be considered
            to be guiding if its separation to the commanded field is larger
            than this value.
        acquisition_timeout
            The maximum time allowed for acquisition. In case of timeout
            the acquired fields are evaluated and an exception is
            raised if the acquisition failed.
        show_progress
            Displays a progress bar with the elapsed exposure time.
        run_cleanup
            Whether to run the cleanup routine.
        adjust_focus
            Adjusts the focuser positions based on temperature drift before
            starting the observation. This works best if the focus has been
            initially determined using a focus sweep.
        cleanup_on_interrupt
            If ``True``, registers a signal handler to catch interrupts and
            run the cleanup routine.

        """

        write_log = self.write_to_log

        # Create tile.
        if isinstance(tile, Tile):
            pass
        elif tile is not None or (tile is None and ra is None and dec is None):
            if use_scheduler:
                tile = await run_in_executor(Tile.from_scheduler, tile_id=tile)
            else:
                raise GortError("Not enough information to create a tile.")

        elif ra is not None and dec is not None:
            if use_scheduler:
                tile = await run_in_executor(Tile.from_scheduler, ra=ra, dec=dec, pa=pa)
            else:
                tile = await run_in_executor(Tile.from_coordinates, ra, dec, pa=pa)

        else:
            raise GortError("Not enough information to create a tile.")

        assert isinstance(tile, Tile)

        if dither_position is not None:
            tile.set_dither_position(dither_position)
        elif tile.sci_coords.dither_position is None:
            self.write_to_log(
                "No dither position defined. Using dither_position=0.",
                "warning",
            )
            tile.set_dither_position(0)

        if cleanup_on_interrupt:
            interrupt_cb = partial(self.gort.run_script_sync, "cleanup")
        else:
            interrupt_cb = None

        is_acquired: bool = False
        is_guiding = self.gort.guiders.sci.status & GuiderStatus.GUIDING

        # We require the tile to be acquired and guiding to skip the slew. The current
        # tile must match the tile_id of the new tile requested and acquisition must
        # have been completed.
        if (
            skip_slew_when_acquired
            and self._tile is not None
            and tile.tile_id
            and self._tile.tile_id == tile.tile_id
            and is_guiding
            and self.stages["acquire"] == ObserverStageStatus.DONE
        ):
            is_acquired = True

        # Reset the tile
        self.reset(tile, on_interrupt=interrupt_cb, reset_stages=not is_acquired)

        # Run the cleanup routine to be extra sure.
        if run_cleanup:
            await self.gort.cleanup(turn_lamps_off=False)

        # Wrap the PA to the range -30 to 30.
        pa = tile.sci_coords.pa
        new_pa = wrap_pa_hex(tile.sci_coords.pa)
        if new_pa != pa:
            write_log(f"Wrapping sci PA from {pa:.3f} to {new_pa:.3f}.")
            tile.sci_coords.pa = new_pa

        if tile.tile_id is not None:
            write_log(
                f"Observing tile_id={tile.tile_id} on "
                f"dither position #{self.dither_position}.",
                "info",
            )

        if adjust_focus:
            await self.gort.guiders.adjust_focus()

        await self.gort.notify_event(
            Event.OBSERVER_NEW_TILE,
            payload={"tile_id": tile.tile_id, "dither_position": self.dither_position},
        )

        exposures: list[Exposure] = []
        failed: bool = False

        try:
            if not is_acquired:
                # Slew telescopes and move fibsel mask.
                await self.slew()

                # Start guiding.
                await self.acquire(
                    guide_tolerance=guide_tolerance,
                    timeout=acquisition_timeout,
                )

            else:
                write_log(f"Acquiring dither position #{self.dither_position}", "info")

                async with GatheringTaskGroup() as group:
                    group.create_task(self.set_dither_position(self.dither_position))
                    if self.standards:
                        group.create_task(self.standards.reacquire_first())

                # Need to restart the guider monitor so that the new exposure
                # gets the range of guider frames that correspond to this dither.
                # GortObserver.expose() doesn't do this because we ask for a single
                # exposure.
                self.guider_monitor.reset()

            # Exposing
            _exposure = await self.expose(
                exposure_time=exposure_time,
                show_progress=show_progress,
                count=n_exposures,
                async_readout=async_readout,
                keep_guiding=keep_guiding,
                dither_position=self.dither_position,
            )

            if not isinstance(_exposure, list):
                exposures = [_exposure]
            else:
                exposures = _exposure

            if self.cancelling:
                await self.gort.guiders.stop()

                write_log("Reading exposure before cancelling.", "warning")
                await asyncio.gather(*exposures)

                raise GortObserverCancelledError()

        except GortObserverCancelledError:
            write_log("Observation cancelled.", "warning")
            failed = len(exposures) == 0

        except KeyboardInterrupt:
            write_log("Observation interrupted by user.", "warning")
            failed = True

        finally:
            # Finish observation.
            await self.finish_observation(keep_guiding=keep_guiding and not failed)

        return (not failed, exposures)

    @handle_signals(interrupt_signals, interrupt_helper.run_callback)
    @register_stage_status
    async def slew(self, telescopes: list[str] = ["sci", "skye", "skyw", "spec"]):
        """Slew to the telescope fields."""

        cotasks = []

        sci: tuple[float, float, float] | None = None
        spec: tuple[float, float] | None = None
        sky: dict[str, tuple[float, float]] = {}

        with self.register_overhead("slew:stop-guiders"):
            # Stops guiders.
            await self.gort.guiders.stop()

        # Slew telescopes.
        self.write_to_log(f"Slewing to tile_id={self.tile.tile_id}.", level="info")

        if "sci" in telescopes:
            sci = (
                self.tile.sci_coords.ra,
                self.tile.sci_coords.dec,
                self.tile.sci_coords.pa,
            )
            self.write_to_log(f"Science: {str(self.tile.sci_coords)}")

        if "spec" in telescopes:
            if self.tile.spec_coords and len(self.tile.spec_coords) > 0:
                first_spec = self.tile.spec_coords[0]

                # For spec we slew to the fibre with which we'll observe first.
                # This should save a bit of time converging.
                spec = fibre_slew_coordinates(
                    first_spec.ra,
                    first_spec.dec,
                    self.mask_positions[0],
                    derotated=False,
                )

                self.write_to_log(f"Spec: {first_spec} on {self.mask_positions[0]}")

        sky = {}
        for skytel in ["SkyE", "SkyW"]:
            skytel_l = skytel.lower()
            if skytel_l in self.tile.sky_coords and skytel_l in telescopes:
                sky_coords_tel = self.tile.sky_coords[skytel_l]
                if sky_coords_tel is not None:
                    sky[skytel_l] = (sky_coords_tel.ra, sky_coords_tel.dec)
                    self.write_to_log(f"{skytel}: {sky_coords_tel}")

        # For sci we want to slew the k-mirror so that we can apply small positive
        # offsets without backlash. So we slew to the tile PA-stop_degs_before.
        kmirror_config = self.gort.config["telescopes"]["kmirror"]
        stop_degs_before = kmirror_config.get("stop_degs_before", {}).get("sci", 0.0)

        cotasks.append(
            self.gort.telescopes.goto(
                sci=sci,
                spec=spec,
                skye=sky.get("skye", None),
                skyw=sky.get("skyw", None),
                sci_km_stop_degs_before=stop_degs_before,
            )
        )

        # Move fibsel to first position.
        fibsel = self.gort.telescopes.spec.fibsel
        cotasks.append(fibsel.move_to_position(self.mask_positions[0], rehome=True))

        # Execute.
        with self.register_overhead("slew:slew"):
            await asyncio.gather(*cotasks)

    @handle_signals(interrupt_signals, interrupt_helper.run_callback)
    @register_stage_status
    async def acquire(
        self,
        guide_tolerance: float | None = None,
        timeout: float = 180,
        telescopes: list[str] = ["sci", "skye", "skyw", "spec"],
    ):
        """Acquires the field in all the telescopes. Blocks until then.

        Parameters
        ----------
        guide_tolerance
            The guide tolerance in arcsec. A telescope will not be considered
            to be guiding if its separation to the commanded field is larger
            than this value. If `None`, default values from the configuration
            file are used.
        timeout
            The maximum time allowed for acquisition. In case of timeout
            the acquired fields are evaluated and an exception is
            raised if the acquisition failed.
        telescopes
            The list of telescopes to acquire. By default all telescopes are
            acquired.

        Raises
        ------
        GortObserverError
            If the acquisition failed or the minimum required telescopes are
            not guiding.

        """

        with self.register_overhead("acquisition:stop-guiders"):
            # Make sure we are not guiding or that the previous is on.
            if self.guide_task is not None and not self.guide_task.done():
                await self.gort.guiders.stop()
                await cancel_task(self.guide_task)

        # Determine telescopes on which to guide.
        guide_coros = []
        guide_on_telescopes: list[str] = []
        n_skies = 0
        for tel in telescopes:
            coords = self.tile[tel]

            if coords is None or (isinstance(coords, list) and len(coords) == 0):
                continue

            if tel == "spec" and isinstance(coords, list):
                coords = coords[0]
            assert isinstance(coords, Coordinates)

            guide_on_telescopes.append(tel)
            if "sky" in tel:
                n_skies += 1

            # Pixel in the MF on which to guide. Always None/central pixel
            # except for sci if defined. For spec we guide on the pixel of the
            # first fibre/mask position.
            pixel = coords._mf_pixel if tel != "spec" else self.mask_positions[0]

            guide_tolerance_tel = self.gort.config["observer"]["guide_tolerance"][tel]
            guide_tolerance_tel = guide_tolerance or guide_tolerance_tel

            guide_coros.append(
                self.gort.guiders[tel].guide(
                    ra=coords.ra,
                    dec=coords.dec,
                    pa=-coords.pa,  # -1 since k-mirror handiness is opposite to tile PA
                    guide_tolerance=guide_tolerance_tel,
                    pixel=pixel,
                )
            )

        if "spec" not in guide_on_telescopes:
            with self.register_overhead("acquisition:move-fibsel-500"):
                self.write_to_log("Not using spec: blocking fibre mask.", "warning")
                await self.gort.telescopes.spec.fibsel.move_relative(500)

        if n_skies == 0:
            self.write_to_log("No sky positions defined.", "warning")

        with self.register_overhead("acquisition:start-guide-loop"):
            # Start guide loop.
            self.guider_monitor.start_monitoring()
            self.guide_task = asyncio.gather(*guide_coros)

        with self.register_overhead("acquisition:acquire"):
            await asyncio.sleep(2)

            # Wait until convergence.
            self.write_to_log("Waiting for guiders to converge.")
            guide_status = await asyncio.gather(
                *[
                    self.gort.guiders[tel].wait_until_guiding(timeout=timeout)
                    for tel in guide_on_telescopes
                ]
            )

        has_timedout = any([gs[3] for gs in guide_status])
        if has_timedout:
            self.write_to_log("Some acquisitions timed out.", "warning")
            # TODO: we need to stop guiders that timed out.

        try:
            for ii, tel in enumerate(guide_on_telescopes):
                is_guiding = guide_status[ii][0]
                if tel == "sci" and not is_guiding:
                    raise GortObserverError(
                        "Science telescope is not guiding.",
                        error_code=ErrorCode.ACQUISITION_FAILED,
                        payload={
                            "observer": True,
                            "tile_id": self.tile.tile_id,
                            "telescope": "sci",
                        },
                    )

                if tel == "spec":
                    if not is_guiding:
                        raise GortObserverError(
                            "Spec telescope is not guiding.",
                            error_code=ErrorCode.ACQUISITION_FAILED,
                            payload={
                                "observer": True,
                                "tile_id": self.tile.tile_id,
                                "telescope": "spec",
                                "fibsel_position": self.mask_positions[0],
                            },
                        )
                    else:
                        await self.gort.guiders.spec.apply_corrections(False)

                if "sky" in tel and not is_guiding:
                    self.write_to_log(f"{tel} telescope is not guiding.", "warning")

        except Exception:
            self.write_to_log("Stopping guide loops.", "warning")
            await self.gort.guiders.stop()

            raise

        self.write_to_log("All telescopes are now guiding.")

    @handle_signals(interrupt_signals, interrupt_helper.run_callback)
    @register_stage_status
    async def expose(
        self,
        exposure_time: float = 900.0,
        show_progress: bool | None = None,
        count: int = 1,
        async_readout: bool = False,
        keep_guiding: bool = True,
        object: str | None = None,
        dither_position: int | None = None,
        exposure_starts_callback: Callable[[Any], Coroutine] | None = None,
    ):
        """Starts exposing the spectrographs.

        Parameters
        ----------
        exposure_time
            The length of the exposure in seconds.
        show_progress
            Displays a progress bar with the elapsed exposure time.
        count
            Number of exposures. If ``iterate_over_standards=True``, a
            full sequence of standards will be observed during each
            exposure.
        async_readout
            Whether to wait for the readout to complete or return as soon
            as the readout begins. If :obj:`False`, the exposure is registered
            but the observation is not finished.
        keep_guiding
            If :obj:`True`, keeps the guider running after the last exposure.
        object
            The object name to be added to the header.
        dither_position
            The dither position. If :obj:`None`, uses the first dither position
            in the tile. Only relevant for exposure registration.
        exposure_starts_callback
            A callback to be called when the exposure starts.

        Returns
        -------
        exposures
            Either a single `.Exposure` or a list of exposures if ``count>1``.

        """

        tile_id = self.tile.tile_id

        dither_position = dither_position or self.tile.sci_coords.dither_position

        if object is None:
            if self.tile.object:
                object = self.tile.object
            elif tile_id is not None and tile_id > 0:
                object = f"tile_id={tile_id}"

        last_exposure = self.gort.specs.last_exposure
        if last_exposure is not None and not last_exposure.done():
            self.write_to_log("Waiting for previous exposure to read out.", "warning")
            await last_exposure

        # Last chance to bail out before the exposure.
        if self.cancelling:
            raise GortObserverCancelledError()

        exposures: list[Exposure] = []
        _n_exposures = 0

        for nexp in range(1, count + 1):
            self.write_to_log(
                f"Starting {exposure_time:.1f} s exposure ({nexp}/{count}).",
                "info",
            )

            with self.register_overhead(f"expose:pre-exposure-{nexp}"):
                # Refresh guider data for this exposure.
                if _n_exposures > 0:
                    self.guider_monitor.reset()

                # Move fibre selector to the first position. Should be there unless
                # count > 1 in which case we need to move it back.
                if self.standards:
                    await self.standards.start_iterating(exposure_time)

                exposure = Exposure(self.gort, flavour="object", object=object)
                self._current_exposure = exposure

            exposure.hooks["pre-readout"].append(self._pre_readout)
            exposure.hooks["post-readout"].append(self._post_readout)

            if exposure_starts_callback:
                exposure.hooks["exposure-starts"].append(exposure_starts_callback)

            with self.register_overhead(f"expose:integration-{nexp}"):
                await exposure.expose(
                    exposure_time=exposure_time,
                    show_progress=show_progress,
                    async_readout=True,
                )

            # TODO: this is a bit dangerous because we are registering the
            # exposure before  it's actually written to disk. Maybe we should
            # wait until _post_readout() to register, but then the scheduler
            # needs to be changed to not return the same tile twice.
            with self.register_overhead(f"exposure:register-exposure-{nexp}"):
                await self.register_exposure(
                    exposure,
                    tile_id=tile_id,
                    dither_position=dither_position,
                )

            if nexp == count and not keep_guiding:
                with self.register_overhead(f"expose:stop-guiders-{nexp}"):
                    await self.gort.guiders.stop()

            if nexp == count and async_readout:
                return exposure
            else:
                await exposure

            exposures.append(exposure)
            _n_exposures += 1

        if len(exposures) == 1:
            return exposures[0]
        else:
            return exposures

    @handle_signals(interrupt_signals, interrupt_helper.run_callback)
    async def finish_observation(self, keep_guiding: bool = False):
        """Finishes the observation, stops the guiders, etc."""

        self.write_to_log("Finishing observation.", "info")

        with self.register_overhead("finish-observation"):
            # Should have been cancelled in _update_header(), but just in case.
            if self.standards is not None:
                await self.standards.cancel()

            if (
                not keep_guiding
                and self.guide_task is not None
                and not self.guide_task.done()
            ):
                await self.gort.guiders.stop()
                await self.guide_task

        tile_id = self.tile.tile_id
        if tile_id is None or tile_id <= 0:
            tile_id = None

        # Write overheads to database.
        payload = [
            {
                "observer_id": id(self),
                "tile_id": tile_id,
                "dither_position": value["dither_position"],
                "stage": name,
                "start_time": value["t0"],
                "end_time": value["t0"] + value["elapsed"],
                "duration": value["elapsed"],
            }
            for name, value in self.overheads.items()
        ]

        try:
            table_name = self.gort.config["services.database.tables.overheads"]
            insert_to_database(table_name, payload)
        except Exception as err:
            self.write_to_log(
                f"Failed saving overheads to database: {decap(err)}",
                "error",
            )

    async def register_exposure(
        self,
        exposure: Exposure,
        tile_id: int | None = None,
        dither_position: int = 0,
    ):
        """Registers the exposure in the database."""

        if exposure._exposure_time is None:
            raise GortObserverError("Exposure time cannot be 'None'.")

        if exposure.flavour != "object":
            return

        standards: list[int] = []
        if self.standards:
            for stdn, std_data in self.standards.standards.items():
                if std_data.observed == 1:
                    pk = self.tile.spec_coords[stdn - 1].pk
                    if pk is not None:
                        standards.append(pk)

        skies = [sky.pk for sky in self.tile.sky_coords.values() if sky.pk is not None]

        self.write_to_log("Registering observation.", "info")
        registration_payload = {
            "dither": dither_position,
            "jd": float(exposure.start_time.jd),
            "seeing": -999.0,
            "standards": standards,
            "skies": skies,
            "exposure_no": exposure.exp_no,
            "exposure_time": exposure._exposure_time,
        }

        if tile_id is not None:
            registration_payload["tile_id"] = tile_id

        self.write_to_log(f"Registration payload {registration_payload}")

        try:
            await register_observation(registration_payload)
        except Exception as err:
            self.write_to_log(f"Failed registering exposure: {decap(err)}", "error")
        else:
            self.write_to_log("Registration complete.")

    async def set_dither_position(self, dither: int):
        """Reacquire science telescope for a new dither position."""

        valid_status = (
            GuiderStatus.GUIDING | GuiderStatus.DRIFTING | GuiderStatus.ACQUIRING
        )

        sci_status = self.gort.guiders.sci.status
        if sci_status is None or not (sci_status & valid_status):
            raise GortObserverError("sci guider must be active to set dither position.")

        self.tile.set_dither_position(dither)
        await self.gort.guiders.sci.set_pixel(self.tile.sci_coords._mf_pixel)

        # Wait until converges.
        with self.register_overhead(f"set-dither-position:acquire-{dither}"):
            await asyncio.sleep(2)
            self.write_to_log("Waiting for 'sci' guider to converge.")
            await self.gort.guiders.sci.wait_until_guiding(timeout=120)

    def write_to_log(
        self,
        message: str,
        level: str = "debug",
        header: str | None = None,
        event: Event | None = None,
        extra_payload: dict[str, Any] = {},
    ):
        """Writes a message to the log with a custom header.

        Parameters
        ----------
        message
            The message to log.
        level
            The level to use for logging: ``'debug'``, ``'info'``, ``'warning'``, or
            ``'error'``.
        header
            The header to prepend to the message. By default uses the class name.
        event
            If specified, emits an event of this type.
        extra_payload
            Additional payload to include in the event.

        """

        if header is None:
            header = f"({self.__class__.__name__}) "

        message = f"{header}{message}"

        level_int = logging._nameToLevel[level.upper()]
        self.gort.log.log(level_int, message)

        if event:
            payload = {
                "observer": True,
                "message": message,
                "level": level,
                "tile_id": self.tile.tile_id,
            }
            payload.update(extra_payload)

            asyncio.create_task(self.gort.notify_event(event, payload=payload))

    @contextmanager
    def register_overhead(self, name: str):
        """Measures and registers and overhead."""

        t0 = time()

        yield

        self.overheads[name] = {
            "dither_position": self.tile.sci_coords.dither_position,
            "t0": t0,
            "elapsed": time() - t0,
        }

    def _get_mask_positions(self, pattern: str):
        """Returns mask positions sorted by motor steps."""

        mask_config = self.gort.config["telescopes"]["mask_positions"]
        all_positions = self.gort.telescopes.spec.fibsel.list_positions()
        positions = [pos for pos in all_positions if re.match(pattern, pos)]

        return sorted(positions, key=lambda p: mask_config[p])

    async def _pre_readout(self, header: dict[str, Any]):
        """Updates the exposure header with pointing and guiding information."""

        tile_id = self.tile.tile_id
        dither_pos = self.tile.sci_coords.dither_position

        # Pointing coordinates taking into account the dither position.
        position_coords = self.tile.sci_coords.get_dither_radec()

        tile_header = {
            "TILE_ID": (tile_id or -999, "The tile_id of this observation"),
            "DPOS": (dither_pos, "Dither position"),
            "POSCIRA": round(position_coords[0], 6),
            "POSCIDE": round(position_coords[1], 6),
            "POSCIPA": round(self.tile.sci_coords.pa, 4),
        }

        if self.tile["skye"]:
            tile_header.update(
                {
                    "POSKYERA": round(self.tile.sky_coords["skye"].ra, 6),
                    "POSKYEDE": round(self.tile.sky_coords["skye"].dec, 6),
                    "POSKYEPA": round(self.tile.sky_coords["skye"].pa, 4),
                    "SKYENAME": self.tile.sky_coords["skye"].name,
                }
            )

        if self.tile["skyw"]:
            tile_header.update(
                {
                    "POSKYWRA": round(self.tile.sky_coords["skyw"].ra, 6),
                    "POSKYWDE": round(self.tile.sky_coords["skyw"].dec, 6),
                    "POSKYWPA": round(self.tile.sky_coords["skyw"].pa, 4),
                    "SKYWNAME": self.tile.sky_coords["skyw"].name,
                }
            )

        header.update(tile_header)

        header.update(self.guider_monitor.to_header())

        # At this point the shutter is closed so let's stop observing standards.
        # This also finishes updating the standards table.
        if self.has_standards and self.standards:
            await self.standards.cancel()

            # There is some overhead between when we set t0 for the first standard
            # and when the exposure actually begins. This leads to the first standard
            # having longer exposure time than open shutter.
            if self.has_standards and self._current_exposure is not None:
                start_time = self._current_exposure.start_time.unix
                self.standards.standards[1].t0 = start_time

            header.update(self.standards.to_header())

    async def _post_readout(self, exposure: Exposure):
        """Post exposure tasks."""

        # tile_id = self.tile.tile_id
        # dither_pos = self.tile.sci_coords.dither_position

        # if exposure.error is True:
        #     if tile_id is not None:
        #         mark_exposure_bad(tile_id, dither_pos)

        #     self.write_to_log(
        #         "Exposure returned with errors. Tile-dither has been marked as bad.",
        #         "error",
        #     )

        return


@dataclass
class Standard:
    """A class to represent a standard star."""

    n: int
    ra: float
    dec: float
    source_id: int = -1
    acquired: bool = False
    observed: bool = False
    t0: float = 0.0
    t1: float = 0.0
    fibre: str = ""


class Standards:
    """Iterates over standards and monitors observed standards."""

    def __init__(
        self,
        observer: GortObserver,
        tile: Tile,
    ):
        self.observer = observer
        self.gort = observer.gort

        self.tile = tile

        self.iterate_task: asyncio.Task | None = None

        self.current_standard: int = 1
        self.standards = self._get_frame()

    @property
    def mask_positions(self):
        """Returns the list of mask positions from `.GortObserver`."""

        return self.observer.mask_positions

    def _get_frame(self):
        """Constructs the standard data frame."""

        standards: dict[int, Standard] = {}

        for stdn, cc in enumerate(self.tile.spec_coords):
            standards[stdn + 1] = Standard(
                n=stdn + 1,
                ra=cc.ra,
                dec=cc.dec,
                source_id=cc.source_id or -1,
            )

        return standards

    async def start_iterating(self, exposure_time: float) -> None:
        """Iterates over the fibre mask positions.

        Moving the ``spec`` telescope to each one of the standard star,
        acquires the new field, and adjusts the time to stay on each target.

        # TODO: right now we are assuming that the first standard has been
        # acquired when we call this method. This is not always the case and
        # it caused a bug when observing multiple dither positions so we should
        # check.

        """

        if len(self.standards) == 0:
            return

        await self.gort.telescopes.spec.fibsel.move_to_position(self.mask_positions[0])
        await cancel_task(self.iterate_task)

        self.standards = self._get_frame()
        self.current_standard = 1

        self.standards[1].acquired = True
        self.standards[1].t0 = time()
        self.standards[1].fibre = self.mask_positions[0]

        self.iterate_task = asyncio.create_task(self._iterate(exposure_time))

    async def reacquire_first(self):
        """Re-acquires the first standard.

        This method should only be called when observing second and subsequent
        dithers in a tile.

        """

        if self.iterate_task and not self.iterate_task.done():
            await self.cancel()

        self.current_standard = 1

        # Home the fibsel just to be sure.
        await self.gort.telescopes.spec.fibsel.home()

        if not (await self.acquire_standard(0)):
            raise GortObserverError("Failed to re-acquire first standard.")

    async def acquire_standard(self, standard_idx: int):
        """Acquires a standard star and starts guiding on it."""

        guider_pixels: dict[str, tuple[float, float]]
        guider_pixels = self.gort.config["guiders"]["devices"]["spec"]["named_pixels"]

        # Tolerance to start guiding
        guide_tolerance_spec = self.gort.config["observer"]["guide_tolerance"]["spec"]

        # Moving the mask to an intermediate position while we move around.
        spec_tel = self.gort.telescopes.spec
        await spec_tel.fibsel.move_relative(500)

        overhead_root = f"standards:standard-{self.current_standard}"

        # New coordinates to observe.
        new_coords = self.tile.spec_coords[standard_idx]
        new_mask_position = self.mask_positions[standard_idx]

        event_payload = {
            "observer": True,
            "tile_id": self.tile.tile_id,
            "telescope": "spec",
            "n_standard": self.current_standard,
            "fibsel_position": new_mask_position,
            "coordinates": [new_coords.ra, new_coords.dec],
        }

        # Pixel on the MF corresponding to the new fibre/mask hole on
        # which to guide. We use this tabulated list instead of
        # offset_to_master_frame_pixel() because the latter coordinates
        # are less precise as they do not include IFU rotation and more
        # precise metrology.
        new_guider_pixel = guider_pixels[new_mask_position]

        self.observer.write_to_log(
            f"Moving to standard #{self.current_standard} ({new_coords}) "
            f"on fibre {new_mask_position}.",
            "info",
        )

        await self.gort.notify_event(
            Event.OBSERVER_ACQUISITION_START,
            payload=event_payload,
        )

        # Slew to new coordinates. We actually slew to the coordinates
        # that make the new star close to the fibre that will observe it.
        slew_ra, slew_dec = fibre_slew_coordinates(
            new_coords.ra,
            new_coords.dec,
            new_mask_position,
            derotated=False,
        )

        # Finish guiding on spec telescope.
        self.observer.write_to_log("Re-slewing 'spec' telescope.")

        with self.observer.register_overhead(f"{overhead_root}-slew"):
            cotasks = [
                self.gort.guiders["spec"].stop(),
                spec_tel.goto_coordinates(ra=slew_ra, dec=slew_dec, force=True),
            ]
            await asyncio.gather(*cotasks)

        # Start to guide. Note that here we use the original coordinates
        # of the star along with the pixel on the master frame on which to
        # guide. See the note in fibre_slew_coordinates().
        self.observer.write_to_log("Starting to guide on spec telescope.")
        asyncio.create_task(
            self.gort.guiders.spec.guide(
                ra=new_coords.ra,
                dec=new_coords.dec,
                guide_tolerance=guide_tolerance_spec,
                pixel=new_guider_pixel,
            )
        )

        with self.observer.register_overhead(f"{overhead_root}-acquire"):
            result = await self.gort.guiders.spec.wait_until_guiding(timeout=60)
            if result[0] is False:
                self.observer.write_to_log(
                    f"Timed out acquiring standard {self.current_standard}.",
                    "warning",
                    event=Event.OBSERVER_STANDARD_ACQUISITION_FAILED,
                    extra_payload={**event_payload, "reason": "timeout"},
                )
                return False

            self.observer.write_to_log(
                f"Standard #{self.current_standard} on "
                f"{new_mask_position!r} has been acquired.",
                "info",
            )

            # Move mask to uncover fibre.
            await spec_tel.fibsel.move_to_position(new_mask_position)

            # Do not guide. This means RA/Dec drifting will happen
            # but not rotation drifting since we are guiding on a point
            # source.
            await self.gort.guiders.spec.apply_corrections(False)

        return True

    async def cancel(self):
        """Cancels iteration."""

        self.observer.write_to_log("Cancelling standards iteration.", "debug")
        await cancel_task(self.iterate_task)

        if len(self.standards) == 0:
            return

        if self.standards[self.current_standard].acquired:
            if not self.standards[self.current_standard].observed:
                self.standards[self.current_standard].observed = True
                self.standards[self.current_standard].t1 = time()

    async def _iterate(self, exposure_time: float):
        """Iterate task."""

        # Time to acquire a standard.
        ACQ_PER_STD = 30

        spec_coords = self.tile.spec_coords

        # If we have zero or one standards, do nothing. The spec telescope
        # is already pointing to the first mask position.
        if len(spec_coords) <= 1:
            return

        # Time at which the exposure began.
        t0 = time()

        # Time to spend on each mask position.
        n_stds = len(spec_coords)

        # Calculate the time to actually be on target, taking into account
        # how long we expect to take acquiring.
        time_per_position = exposure_time / n_stds - ACQ_PER_STD
        if time_per_position < 0:
            self.observer.write_to_log(
                "Exposure time is too short to observe this "
                "many standards. I will do what I can.",
                "warning",
            )
            time_per_position = exposure_time / n_stds

        # Time at which we started observing the last standard.
        t0_last_std = t0

        # Number of standards observed.
        n_observed = 1

        # Index of current standard being observed.
        current_std_idx = 0

        while True:
            await asyncio.sleep(1)

            # We consider than if there is less 2 * ACQ_PER_STD left in the
            # exposure there is no point in going to the next standard.
            t_now = time()
            if t_now - t0 > exposure_time - 2 * ACQ_PER_STD:
                self.observer.write_to_log("Exiting standard loop.")
                self.observer.write_to_log(
                    f"Standards observed: {n_observed}/{n_stds}.",
                    "info",
                )
                return

            # Time to move to another standard?
            if t_now - t0_last_std > time_per_position:
                # Check that we haven't run out standards. If so,
                # keep observing this one.
                if len(spec_coords) == current_std_idx + 1:
                    continue

                # Register the previous standard.
                if self.standards[self.current_standard].acquired:
                    self.standards[self.current_standard].t1 = time()
                    self.standards[self.current_standard].observed = True

                # Increase current index and acquire the next standard.
                current_std_idx += 1
                self.current_standard += 1

                if not (await self.acquire_standard(current_std_idx)):
                    continue

                n_observed += 1
                t0_last_std = time()

                new_mask_position = self.mask_positions[current_std_idx]

                self.standards[self.current_standard].acquired = True
                self.standards[self.current_standard].t0 = time()
                self.standards[self.current_standard].fibre = new_mask_position

    def to_header(self):
        """Returns observed standards as a header-ready dictionary."""

        header_data = {}

        for nstd, data in self.standards.items():
            header_data[f"STD{nstd}ID"] = data.source_id if data.source_id > 0 else None
            header_data[f"STD{nstd}RA"] = data.ra
            header_data[f"STD{nstd}DE"] = data.dec
            header_data[f"STD{nstd}ACQ"] = data.observed

            if data.observed:
                header_data[f"STD{nstd}T0"] = Time(data.t0, format="unix").isot
                header_data[f"STD{nstd}T1"] = Time(data.t1, format="unix").isot
                header_data[f"STD{nstd}EXP"] = round(data.t1 - data.t0, 1)
                header_data[f"STD{nstd}FIB"] = data.fibre

        return header_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-10-27
# @Filename: dome.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum

from typing import TYPE_CHECKING, Coroutine

from lvmopstools.retrier import Retrier

from gort.exceptions import GortError


if TYPE_CHECKING:
    from gort.overwatcher.overwatcher import Overwatcher


__all__ = ["DomeHelper", "DomeStatus"]


class DomeStatus(enum.Flag):
    """An enumeration of dome statuses."""

    OPEN = enum.auto()
    CLOSED = enum.auto()
    OPENING = enum.auto()
    CLOSING = enum.auto()
    MOVING = enum.auto()
    UNKNOWN = enum.auto()


class DomeHelper:
    """Handle dome movement."""

    def __init__(self, overwatcher: Overwatcher):
        self.overwatcher = overwatcher
        self.gort = overwatcher.gort
        self.log = overwatcher.log

        self._move_lock = asyncio.Lock()
        self.locked: bool = False  # Prevents the dome from operating.

    def reset(self):
        """Resets the dome lock."""

        self.locked = False

    @Retrier(max_attempts=3, delay=1)
    async def status(self, force: bool = False):
        """Returns the status of the dome.

        Parameters
        ----------
        force
            If :obj:`True`, will force the status check by issuing a command to
            the ``lvmecp`` actor. If :obj:`False` and the ``lvmecp`` actor model
            is being tracked by the `.Gort` instance, uses the last seen value.

        """

        if force:
            status = await self.gort.enclosure.status()
            labels = status["dome_status_labels"]
        else:
            try:
                labels = self.gort.models["lvmecp"]["dome_status_labels"].value
                if labels is None:
                    raise ValueError("dome_status_labels is None.")
            except Exception:
                return await self.status(force=True)

        labels = labels.split(",")

        if "MOTOR_OPENING" in labels:
            return DomeStatus.OPENING | DomeStatus.MOVING
        elif "MOTOR_CLOSING" in labels:
            return DomeStatus.CLOSING | DomeStatus.MOVING
        elif "OPEN" in labels:
            if "MOVING" in labels:  # This probably does not happen.
                return DomeStatus.OPENING | DomeStatus.MOVING
            return DomeStatus.OPEN
        elif "CLOSED" in labels:
            if "MOVING" in labels:  # This probably does not happen.
                return DomeStatus.CLOSING | DomeStatus.MOVING
            return DomeStatus.CLOSED

        return DomeStatus.UNKNOWN

    async def is_opening(self):
        """Returns True if the dome is open or opening."""

        status = await self.status()

        if status & DomeStatus.OPENING:
            return True
        if status & DomeStatus.OPEN:
            return True

        return False

    async def is_closed(self):
        """Returns True if the dome is closed."""

        status = await self.status()

        if status & DomeStatus.CLOSED:
            return True

        return False

    async def is_closing(self):
        """Returns True if the dome is closed or closing."""

        status = await self.status()

        if status & DomeStatus.CLOSING:
            return True
        if status & DomeStatus.CLOSED:
            return True

        return False

    async def is_moving(self):
        """Returns True if the dome is moving."""

        status = await self.status()

        if status & DomeStatus.MOVING:
            return True

        return False

    async def wait_until_idle(self, timeout: float | None = None):
        """Waits until the dome is idle."""

        elapsed: float = 0

        while await self.is_moving():
            await asyncio.sleep(2)
            elapsed += 2

            if timeout and elapsed >= timeout:
                raise GortError("Timed out waiting for the dome to become idle.")

    async def _move_with_retries(
        self,
        open: bool = False,
        park: bool = True,
        retry_on_close: bool = True,
    ):
        """Moves the dome with retries."""

        try:
            if open:
                await self.gort.enclosure.open(park_telescopes=park)
            else:
                await self.gort.enclosure.close(park_telescopes=park)

        except Exception as err:
            if open or not retry_on_close:
                # Do not retry if opening failed.
                raise

            # If closing, try a second time with the overcurrent mode
            # and without parking the telescopes.
            await self.overwatcher.notify(
                "The dome has failed to close. Retrying in overcurrent mode. "
                "The original error was:",
                level="error",
                error=err,
            )

            await asyncio.sleep(5)  # Wait a bit to give the PLC time to clear.
            await self.gort.enclosure.close(
                park_telescopes=park,
                # With force=True will try to park the telescopes
                # if park=True but will try to close if parking fails.
                force=True,
                mode="overcurrent",
            )

    async def _move(
        self,
        status: DomeStatus,
        open: bool = False,
        park: bool = True,
        retry_on_close: bool = True,
    ):
        """Moves the dome."""

        if self.locked:
            raise GortError("Dome is locked. Cannot open/close.")

        if open and not self.overwatcher.state.safe:
            raise GortError("Cannot open the dome when conditions are unsafe.")

        if not await self.gort.enclosure.allowed_to_move():
            raise GortError("Dome found in invalid or unsafe state.")

        try:
            if status & DomeStatus.MOVING:
                self.log.warning("Dome is already moving. Stopping.")
                await self.stop()

            await self._move_with_retries(
                open=open,
                park=park,
                retry_on_close=retry_on_close,
            )

        except Exception:
            await asyncio.sleep(3)

            status = await self.status()

            if status & DomeStatus.MOVING:
                self.log.warning("Dome is still moving after an error. Stopping.")
                await self.stop()

            # Sometimes the open/close could fail but actually the dome is open/closed.
            if open and not (status & DomeStatus.OPEN):
                raise GortError("Dome is not open after a move command.")
            elif not open and not (status & DomeStatus.CLOSED):
                raise GortError("Dome is not closed after a move command.")

    async def _run_or_disable(self, coro: Coroutine):
        """Runs a coroutine or disables the overwatcher if it fails."""

        try:
            await coro
        except Exception as err:
            await self.overwatcher.notify(
                "The dome has failed to open/close. Disabling the Overwatcher "
                "to prevent further attempts. Please check the dome immediately, "
                "it may be partially or fully open.",
                level="critical",
                error=err,
            )

            # Release the lock here.
            if self._move_lock.locked():
                self._move_lock.release()

            await self.overwatcher.shutdown(
                "Dome failed to open/close. Disabling the overwatcher "
                "and locking the dome. No further attempts will be made to "
                "open/close until the dome lock is reset.",
                disable_overwatcher=True,
                close_dome=False,
            )
            self.locked = True

            raise

    async def open(self, park: bool = True):
        """Opens the dome.

        Parameters
        ----------
        park
            Whether to park the telescopes before opening the dome.

        """

        async with self._move_lock:
            status = await self.status()
            if status == DomeStatus.OPEN:
                self.log.debug("Dome is already open.")
                return

            if status == DomeStatus.OPENING:
                self.log.debug("Dome is already opening. Waiting ...")
                await self.wait_until_idle(timeout=250)
                return

            if status == DomeStatus.UNKNOWN:
                self.log.warning("Dome is in an unknown status. Stopping and opening.")
                await self.stop()
                status = await self.status()

            await self._run_or_disable(self._move(status, open=True, park=park))

    async def close(self, park: bool = True, retry: bool = True):
        """Closes the dome.

        Parameters
        ----------
        park
            Whether to park the telescopes before closing the dome.
        retry
            If :obj:`True`, retries closing the dome in overcurrent mode if the first
            attempt fails.

        """

        async with self._move_lock:
            status = await self.status()
            if status == DomeStatus.CLOSED:
                self.log.debug("Dome is already closed.")
                return

            if status == DomeStatus.CLOSING:
                self.log.debug("Dome is already closing. Waiting ...")
                await self.wait_until_idle(timeout=250)

            if status == DomeStatus.UNKNOWN:
                self.log.warning("Dome is in an unknown status. Stopping and closing.")
                await self.stop()
                status = await self.status()

            await self._run_or_disable(
                self._move(
                    status,
                    open=False,
                    park=park,
                    retry_on_close=retry,
                )
            )

    async def stop(self):
        """Stops the dome."""

        await self.gort.enclosure.stop()

        await asyncio.sleep(1)
        status = await self.status()

        if status & DomeStatus.MOVING:
            raise GortError("Dome is still moving after a stop command.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-26
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


from __future__ import annotations

from .alerts import AlertsOverwatcher
from .calibrations import CalibrationsOverwatcher
from .core import OverwatcherModule
from .ephemeris import EphemerisOverwatcher
from .events import EventsOverwatcher
from .health import HealthOverwatcher
from .helpers import DomeHelper, DomeStatus, NotifierMixIn
from .observer import ObserverOverwatcher
from .overwatcher import Overwatcher
from .safety import SafetyOverwatcher
from .transparency import TransparencyOverwatcher



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-08-13
# @Filename: calibrations.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import random

from typing import Sequence

import numpy
from astropy.time import Time

from gort.enums import ErrorCode, Event
from gort.exceptions import GortError
from gort.tools import decap, get_ephemeris_summary

from .base import BaseRecipe


__all__ = ["QuickCals", "BiasSequence", "TwilightFlats", "LongTermCalibrations"]


class ExposureTimeTooLong(GortError):
    """Raised when the exposure time is too long to continue twilight flats."""

    pass


class QuickCals(BaseRecipe):
    """Runs a quick calibration sequence."""

    name = "quick_cals"

    async def recipe(self):
        """Runs the calibration sequence."""

        specs_idle = await self.gort.specs.are_idle()
        if not specs_idle:
            raise RuntimeError("Spectrographs are not idle.")

        await self.gort.cleanup()

        self.gort.log.info("Pointing telescopes to the calibration screen.")
        await self.gort.telescopes.goto_named_position("calibration")

        ########################
        # Arcs
        ########################

        self.gort.log.info("Turning on the HgNe lamp.")
        await self.gort.nps.calib.on("HgNe")

        self.gort.log.info("Turning on the Ne lamp.")
        await self.gort.nps.calib.on("Neon")

        self.gort.log.info("Turning on the Argon lamp.")
        await self.gort.nps.calib.on("Argon")

        self.gort.log.info("Turning on the Xenon lamp.")
        await self.gort.nps.calib.on("Xenon")

        self.gort.log.info("Waiting 180 seconds for the lamps to warm up.")
        await asyncio.sleep(180)

        fiber = random.randint(1, 12)  # select random fibre on std telescope
        fiber_str = f"P1-{fiber}"
        self.gort.log.info(f"Taking {fiber_str} exposure.")
        await self.gort.telescopes.spec.fibsel.move_to_position(fiber_str)

        for exp_time in [10, 50]:
            await self.gort.specs.expose(
                exp_time,
                flavour="arc",
                header={"CALIBFIB": f"P1-{fiber}"},
            )

        self.gort.log.info("Turning off all lamps.")
        await self.gort.nps.calib.all_off()

        ########################
        # Flats
        ########################

        self.gort.log.info("Turning on the Quartz lamp.")
        await self.gort.nps.calib.on("Quartz")

        self.gort.log.info("Waiting 120 seconds for the lamp to warm up.")
        await asyncio.sleep(120)

        exp_quartz = 20
        await self.gort.specs.expose(
            exp_quartz,
            flavour="flat",
            header={"CALIBFIB": f"P1-{fiber}"},
        )

        self.gort.log.info("Turning off the Quartz lamp.")
        await self.gort.nps.calib.all_off()

        self.gort.log.info("Turning on the LDLS lamp.")
        await self.gort.nps.calib.on("LDLS")

        self.gort.log.info("Waiting 300 seconds for the lamp to warm up.")
        await asyncio.sleep(300)

        exp_LDLS = 150

        await self.gort.specs.expose(
            exp_LDLS,
            flavour="flat",
            header={"CALIBFIB": f"P1-{fiber}"},
        )

        self.gort.log.info("Turning off the LDLS lamp.")
        await self.gort.nps.calib.all_off()


class BiasSequence(BaseRecipe):
    """Takes a sequence of bias frames."""

    name = "bias_sequence"

    async def recipe(self, count: int = 7):
        """Takes a sequence of bias frames.

        Parameters
        ----------
        count
            The number of bias frames to take.

        """

        specs_idle = await self.gort.specs.are_idle()
        if not specs_idle:
            raise RuntimeError("Spectrographs are not idle.")

        self.gort.log.info("Pointing telescopes to the selfie position.")
        await self.gort.telescopes.goto_named_position("selfie")

        await self.gort.nps.calib.all_off()
        await self.gort.cleanup()

        for _ in range(count):
            await self.gort.specs.expose(flavour="bias")


class TwilightFlats(BaseRecipe):
    """Takes a sequence of twilight flats."""

    name = "twilight_flats"

    async def recipe(
        self,
        wait: bool = True,
        start_fibre: int | None = None,
        secondary: bool = False,
    ):
        """Takes a sequence of twilight flats.

        Based on K. Kreckel's code.

        """

        config = self.gort.config["recipes"][self.name]

        # Exposure time model
        popt: numpy.ndarray = numpy.array(config["popt"])

        # Start sunset flats two minutes after sunset.
        # Positive numbers means "into" the twilight.
        sunset_start: float = config.get("sunset_start", 2)

        # Start sunrise flats 15 minutes before sunrise
        sunrise_start: float = config.get("sunrise_start", 15)

        # Fudge factor for the exposure time, in minutes.
        fudge_factor: float = config.get("fudge_factor", 0)

        # Minimum exposure time for normal flats.
        min_exp_time: float = config.get("min_exp_time", 1)

        # Maximum exposure time for normal flats.
        max_exp_time: float = config.get("max_exp_time", 300)

        # Maximum exposure time for extra flats.
        max_exp_time_extra: float = config.get("max_exp_time_extra", 100)

        # Whether to clip the exposure time to the maximum.
        clip_to_max_exposure_time: bool = config.get("clip_to_max_exposure_time", True)

        await self.gort.cleanup()

        if not (await self.gort.enclosure.is_open()):
            raise RuntimeError("Dome must be open to take twilight flats.")

        specs_idle = await self.gort.specs.are_idle()
        if not specs_idle:
            raise RuntimeError("Spectrographs are not idle.")

        eph = await get_ephemeris_summary()

        is_sunset: bool = False
        is_sunrise: bool = False

        if abs(eph["time_to_sunset"]) < abs(eph["time_to_sunrise"]):
            is_sunset = True
            riseset = Time(eph["sunset"], format="jd", scale="utc")
            alt = 40.0
            az = 270.0
        else:
            is_sunrise = True
            riseset = Time(eph["sunrise"], format="jd", scale="utc")
            alt = 40.0
            az = 90.0

        self.gort.log.info("Moving telescopes to point to the twilight sky.")
        await self.gort.telescopes.goto_coordinates_all(
            alt=alt,
            az=az,
            altaz_tracking=False,
        )

        n_fibre = start_fibre or random.randint(1, 12)
        await self.goto_fibre_position(n_fibre, secondary=secondary)

        n_observed: int = 0
        total_fibres: int = 12
        all_done: bool = False

        while True:
            # Calculate the number of minutes into the twilight. Positive values
            # mean minutes into daytime (before sunset or after sunrise).
            now = Time.now()
            time_diff_sun = (now - riseset).sec / 60.0  # Minutes
            if is_sunset:
                time_diff_sun = -time_diff_sun

            # Calculate exposure time.
            aa, bb, cc = popt
            exp_time = aa * numpy.exp(-(time_diff_sun + fudge_factor) / bb) + cc

            if is_sunset:
                time_to_flat_twilighs = sunset_start + time_diff_sun
            else:
                time_to_flat_twilighs = -(sunrise_start + time_diff_sun)

            if time_to_flat_twilighs > 0:
                if wait:
                    self.gort.log.info(
                        "Waiting for twilight. Time to twilight flats: "
                        f"{time_to_flat_twilighs:.1f} minutes."
                    )
                    await asyncio.sleep(time_to_flat_twilighs * 60)
                    continue
                else:
                    raise RuntimeError("Too early to take twilight flats.")

            if not all_done and exp_time < max_exp_time:
                # We allow negative times, which will be rounded to 1 s
                pass
            elif is_sunrise and exp_time > 400 and wait:
                self.gort.log.info(f"Exposure time is too long ({exp_time:.1f} s).")
                self.gort.log.info("Waiting 10 seconds ...")
                await asyncio.sleep(10)
                continue
            elif is_sunset and not all_done and exp_time > max_exp_time:
                if clip_to_max_exposure_time:
                    self.gort.log.info(
                        f"Exposure time is too long ({exp_time:.1f} s). "
                        f"Clipping to {max_exp_time} seconds."
                    )
                    exp_time = max_exp_time
                else:
                    # We have not yet taken all 12 fibres but the
                    # exposure time  is just too long.
                    raise ExposureTimeTooLong(
                        f"Exposure time is now too long to continue taking flats. "
                        f"Took flats for {n_observed}/{total_fibres} fibres."
                    )
            elif is_sunset and all_done and exp_time > max_exp_time_extra:
                # We have taken all 12 fibres and some extra ones, but the
                # exposure time is now too long. Exit.
                self.gort.log.debug("Exposure time is too long to take more flats.")
                break
            elif is_sunset and all_done and exp_time <= max_exp_time_extra:
                # Continue taking exposures for extra flats.
                pass
            else:
                raise RuntimeError("Too early/late to take twilight flats.")

            # Round to the nearest second.
            exp_time = numpy.round(exp_time, 1)

            # Require a minimum exposure time.
            if exp_time < min_exp_time:
                await asyncio.sleep(2)
                continue

            if exp_time < 1:
                exp_time = 1.0

            fibre_str = await self.goto_fibre_position(n_fibre, secondary=secondary)
            self.gort.log.info(f"Exposing {fibre_str} with exp_time={exp_time:.2f}.")

            try:
                await self.gort.specs.expose(
                    float(exp_time),
                    flavour="flat",
                    header={"CALIBFIB": fibre_str},
                )
            except Exception as err:
                self.gort.log.error(
                    "Error taking twilight flat exposure. Will ignore since we "
                    f"are on a schedule here. Error is: {decap(err)}"
                )
                await self.gort.notify_event(
                    Event.ERROR,
                    payload={
                        "error": str(err),
                        "error_code": ErrorCode.CALIBRATION_ERROR.value,
                        "ignore": True,
                    },
                )

            n_observed += 1
            if n_observed == total_fibres:
                all_done = True

                if (all_done and is_sunrise) or exp_time >= max_exp_time_extra:
                    break

                # During sunset, after taking all 12 fibres we continue taking
                # flats until the exposure time reaches MAX_EXP_TIME_EXTRA seconds.
                self.gort.log.info(
                    f"All fibres observed. Taking extra flats until the "
                    f"exposure time reaches {max_exp_time_extra} seconds."
                )

            n_fibre = n_fibre + 1
            if n_fibre > 12:
                n_fibre -= 12

    async def goto_fibre_position(self, n_fibre: int, secondary: bool = False):
        """Moves the mask to a fibre position."""

        fibre_str = f"P1-{n_fibre}"
        if secondary:
            fibre_str = f"P2-{n_fibre}"
        await self.gort.telescopes.spec.fibsel.move_to_position(fibre_str)

        return fibre_str


class LongTermCalibrations(BaseRecipe):
    """Runs the long-term calibration sequence."""

    name = "long_term_calibrations"

    async def recipe(self):
        """Runs the calibration sequence."""

        config = self.gort.config["recipes"][self.name]

        n_biases: int = config.get("n_biases", 7)
        quartz_exp_time: float = config.get("quartz_exp_time", 20)
        ldls_exp_time: float = config.get("ldls_exp_time", 150)

        arc_exp_times: Sequence[float] = config.get("arc_exp_times", [10, 50])
        if not isinstance(arc_exp_times, Sequence):
            arc_exp_times = [arc_exp_times]

        self.gort.log.info("Moving telescopes to point to the calibration screen.")
        await self.gort.telescopes.goto_named_position("calibration")

        # Quart flats

        self.gort.log.info("Turning on the Quartz lamp.")
        await self.gort.nps.calib.on("Quartz")

        self.gort.log.info("Waiting 120 seconds for the lamp to warm up.")
        await asyncio.sleep(120)

        for fibre in range(12):
            fibre_str = f"P1-{fibre + 1}"
            self.gort.log.info(f"Taking fibre {fibre_str} exposure.")

            await self.gort.telescopes.spec.fibsel.move_to_position(fibre_str)
            await self.gort.specs.expose(
                quartz_exp_time,
                flavour="flat",
                header={"CALIBFIB": fibre_str},
            )

        self.gort.log.debug("Turning off the Quartz lamp.")
        await self.gort.nps.calib.all_off()

        # LDLS flats

        self.gort.log.info("Turning on the LDLS lamp.")
        await self.gort.nps.calib.on("LDLS")

        self.gort.log.info("Waiting 300 seconds for the lamp to warm up.")
        await asyncio.sleep(300)

        for fibre in range(12):
            fibre_str = f"P1-{fibre + 1}"
            self.gort.log.info(f"Taking fibre {fibre_str} exposure.")

            await self.gort.telescopes.spec.fibsel.move_to_position(fibre_str)
            await self.gort.specs.expose(
                ldls_exp_time,
                flavour="flat",
                header={"CALIBFIB": fibre_str},
            )

        self.gort.log.debug("Turning off the LDLS lamp.")
        await self.gort.nps.calib.all_off()

        # Arcs
        self.gort.log.info("Turning on arc lamps.")

        await self.gort.nps.calib.on("HgNe")
        await self.gort.nps.calib.on("Neon")
        await self.gort.nps.calib.on("Argon")
        await self.gort.nps.calib.on("Xenon")

        self.gort.log.info("Waiting 180 seconds for the lamps to warm up.")
        await asyncio.sleep(180)

        for fibre in range(12):
            fibre_str = f"P1-{fibre + 1}"
            self.gort.log.info(f"Taking fibre {fibre_str} exposures.")

            await self.gort.telescopes.spec.fibsel.move_to_position(fibre_str)
            for arc_exp_time in arc_exp_times:
                await self.gort.specs.expose(
                    arc_exp_time,
                    flavour="arc",
                    header={"CALIBFIB": fibre_str},
                )

        self.gort.log.debug("Turning off all lamps.")
        await self.gort.nps.calib.all_off()

        # Biases
        await self.gort.execute_recipe("bias_sequence", count=n_biases)

        # Park the telescopes after the calibrations.
        await self.gort.telescopes.park(disable=False)

        self.gort.log.info("Long-term calibrations completed successfully.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-26
# @Filename: ephemeris.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import pathlib
from time import time

from typing import Literal

from astropy.time import Time
from pydantic import BaseModel

from sdsstools import get_sjd

from gort.overwatcher import OverwatcherModule
from gort.overwatcher.core import OverwatcherModuleTask
from gort.tools import decap, get_ephemeris_summary


__all__ = ["EphemerisOverwatcher"]


class EphemerisModel(BaseModel):
    SJD: int
    request_jd: float
    date: str
    sunset: float
    twilight_end: float
    twilight_start: float
    sunrise: float
    is_night: bool
    is_twilight: bool
    time_to_sunset: float
    time_to_sunrise: float
    moon_illumination: float
    from_file: bool


class EphemerisMonitorTask(OverwatcherModuleTask["EphemerisOverwatcher"]):
    """Monitors the ephemeris."""

    name = "ephemeris_monitor"
    keep_alive = True
    restart_on_error = True

    async def task(self):
        """Monitors SJD change and keeps ephemeris updated."""

        failing: bool = False

        while True:
            sjd = get_sjd("LCO")

            if self.module.ephemeris is None or failing or sjd != self.module.sjd:
                try:
                    # First, roll over the GORT log to the new SJD file.
                    log = self.gort.log
                    if log.fh and log.log_filename:
                        path = pathlib.Path(log.log_filename)
                        if str(sjd) not in path.name:
                            log.fh.flush()
                            log.fh.close()
                            log.removeHandler(log.fh)
                            log.start_file_logger(str(path.parent / f"{sjd}.log"))

                    await self.module.update_ephemeris(sjd)

                except Exception as err:
                    await self.notify(
                        f"Failed getting ephemeris data for {sjd}: {decap(err)}",
                        level="error",
                    )
                    failing = True

                else:
                    self.log.info(f"New SJD: updating ephemeris for {sjd}.")
                    self.module.sjd = sjd

                    failing = False

            # Check the calibrations SJD and reset if necessary.
            if self.overwatcher.calibrations.schedule.sjd != sjd:
                await self.overwatcher.calibrations.reset()
                self.overwatcher.troubleshooter.reset(clear_all_tracking=True)

                # Ensure that the Overwatcher is disabled.
                await self.overwatcher.shutdown(
                    "Disabling overwatcher on SJD change.",
                    close_dome=False,
                    disable_overwatcher=True,
                    park=False,
                )

            await asyncio.sleep(60)


class EphemerisOverwatcher(OverwatcherModule):
    """Monitors ephemeris data."""

    name = "ephemeris"

    tasks = [EphemerisMonitorTask()]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.config = self.overwatcher.config

        self.sjd = get_sjd("LCO")
        self.ephemeris: EphemerisModel | None = None
        self.last_updated: float = 0.0

    async def update_ephemeris(self, sjd: int | None = None):
        """Updates the ephemeris data."""

        sjd = sjd or get_sjd("LCO")

        ephemeris_response = await get_ephemeris_summary(sjd)
        self.last_updated = time()

        self.ephemeris = EphemerisModel(**ephemeris_response)

        return self.ephemeris

    def is_night(self, mode: Literal["strict", "observer"] = "strict"):
        """Determines whether it is night-time.

        Parameters
        ----------
        mode
            The mode to use. If ``'strict'``, the function will return :obj:`True`
            if the current time is between 15-degree twilights. If ``'observer'``,
            it will also return :obj:`False` if we are within
            ``scheduler.stop_secs_before_morning`` seconds of the morning twilight.
            In this mode it also returns :obj:`True`
            ``scheduler.open_dome_secs_before_twilight`` seconds before the evening
            twilight to allow time to open the dome before starting observations.

        """

        ephemeris = self.ephemeris
        now_jd = float(Time.now().jd)

        schedule_config = self.config["overwatcher.schedule"]
        stop_secs = schedule_config["stop_secs_before_morning"]
        open_dome_secs = schedule_config["open_dome_secs_before_twilight"]

        if not ephemeris:
            self.log.warning("Ephemeris data not available. is_night() returns False.")
            return False

        twilight_end = float(ephemeris.twilight_end)
        twilight_start = float(ephemeris.twilight_start)
        if mode == "observer":
            twilight_end -= open_dome_secs / 86400
            twilight_start -= stop_secs / 86400

        return twilight_end < now_jd < twilight_start

    def time_to_evening_twilight(self):
        """Returns the time to evening twilight in seconds."""

        ephemeris = self.ephemeris
        if not ephemeris:
            return None

        now = time()
        evening_twilight = Time(ephemeris.twilight_end, format="jd").unix

        return round(evening_twilight - now, 2)

    def time_to_morning_twilight(self, mode: Literal["strict", "observer"] = "strict"):
        """Returns the time to morning twilight in seconds."""

        ephemeris = self.ephemeris
        if not ephemeris:
            return None

        now = time()

        twilight_time = Time(ephemeris.twilight_start, format="jd").unix
        if mode == "observer":
            stop_secs = self.config["overwatcher.schedule.stop_secs_before_morning"]
            twilight_time -= stop_secs

        return round(twilight_time - now, 2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-25
# @Filename: ag.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from gort.devices.core import GortDevice, GortDeviceSet
from gort.gort import Gort
from gort.tools import run_lvmapi_task


__all__ = ["AG", "AGSet"]


class AG(GortDevice):
    """Class representing an AG camera."""

    def __init__(self, gort: Gort, name: str, actor: str, **device_data):
        super().__init__(gort, name, actor)

        self.telescope = name
        self.ips = {
            "east": device_data["ips"]["east"]
            if "ips" in device_data and "east" in device_data["ips"]
            else None,
            "west": device_data["ips"]["west"]
            if "ips" in device_data and "west" in device_data["ips"]
            else None,
        }

    @property
    def n_cameras(self):
        """The number of AG cameras for this telescope."""

        return len([1 for ip in self.ips.values() if ip is not None])

    async def status(self):
        """Returns the status of the AG."""

        return await self.actor.commands.status()

    async def is_idle(self):
        """Returns :obj:`True` if all the cameras are idle."""

        status = await self.status()
        for reply in status.replies:
            if reply["status"]["camera_state"] != "idle":
                return False

        return True

    async def reconnect(self):
        """Reconnect the AG cameras."""

        self.write_to_log("Reconnecting AG cameras.")

        return await self.actor.commands.reconnect()

    async def expose(
        self,
        exposure_time: float = 5.0,
        flavour: str = "object",
        **kwargs,
    ):
        """Exposes the cameras.

        Parameters
        ----------
        exposure_time
            The amount of time, in seconds, to expose the cameras.
        flavour
            The type of image to take, one of ``'bias'``, ``'dark'``, ``'flat'``,
            or ``'object'``.
        kwargs
            Any parameters to send to the ``lvmcam expose`` command.

        """

        if flavour == "bias":
            kwargs["bias"] = True
        elif flavour == "flat":
            kwargs["flat"] = True
        elif flavour == "dark":
            kwargs["dark"] = True
        else:
            kwargs["object"] = True

        return await self.actor.commands.expose(
            exptime=exposure_time,
            **kwargs,
        )


class AGSet(GortDeviceSet[AG]):
    """A set of auto-guiders."""

    __DEVICE_CLASS__ = AG
    __DEPLOYMENTS__ = ["lvmagcam"]

    @property
    def n_cameras(self):
        """The number of AG cameras in the set."""

        return sum([ag.n_cameras for ag in self.values()])

    async def reconnect(self):
        """Reconnects all the AG cameras.

        Returns
        -------
        cameras
            A list of cameras available after reconnecting.

        """

        await asyncio.gather(*[ag.reconnect() for ag in self.values()])

        return await self.list_alive_cameras()

    async def are_idle(self):
        """Returns :obj:`True` if all the cameras are idle."""

        return all(await asyncio.gather(*[ag.is_idle() for ag in self.values()]))

    async def list_alive_cameras(self):
        """Returns a list of cameras found alive and well.

        Currently when a camera has disconnected, the status command still reports
        it as "online", but it doesn't report its temperature or other parameters.

        """

        all_status = await asyncio.gather(*[ag.status() for ag in self.values()])

        replies = []
        for status in all_status:
            for reply in status.replies:
                try:
                    replies.append({"actor": status.actor.name, **reply["status"]})
                except Exception:
                    pass

        cameras = []
        for reply in replies:
            actor = reply["actor"]
            telescope = actor.split(".")[1]
            camera = reply.get("camera", None)
            if camera:
                cameras.append(f"{telescope}-{camera}")

        return cameras

    async def expose(
        self,
        exposure_time: float = 5.0,
        flavour: str = "object",
        **kwargs,
    ):
        """Exposes the cameras.

        Parameters
        ----------
        exposure_time
            The amount of time, in seconds, to expose the cameras.
        flavour
            The type of image to take, one of ``'bias'``, ``'dark'``, ``'flat'``,
            or ``'object'``.
        kwargs
            Any parameters to send to the ``lvmcam expose`` command.

        """

        return await asyncio.gather(
            *[
                ag.expose(
                    exposure_time=exposure_time,
                    flavour=flavour,
                    **kwargs,
                )
                for ag in self.values()
            ]
        )

    async def power_cycle(self):
        """Power cycles all the cameras."""

        await run_lvmapi_task("/macros/power_cycle_ag_cameras")
        await asyncio.sleep(30)

        for retry in range(2):
            try:
                alive_cameras = await self.list_alive_cameras()
            except Exception:
                alive_cameras = []

            if len(alive_cameras) != self.n_cameras:
                if retry == 1:
                    raise RuntimeError("Not all cameras are responding.")

                self.write_to_log(
                    "Not all cameras are responding. Waiting 30 seconds and retrying.",
                    "warning",
                )
                await asyncio.sleep(30)
                await self.reconnect()

        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2024-03-27
# @Filename: observer.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from time import time

from typing import TYPE_CHECKING

from lvmopstools.retrier import Retrier

from gort.exceptions import (
    GortError,
    TroubleshooterCriticalError,
    TroubleshooterTimeoutError,
)
from gort.exposure import Exposure
from gort.overwatcher import OverwatcherModule
from gort.overwatcher.core import OverwatcherModuleTask
from gort.overwatcher.transparency import TransparencyQuality
from gort.overwatcher.troubleshooter.recipes import AcquisitionFailedRecipe
from gort.tile import Tile
from gort.tools import cancel_task, decap, ensure_period, run_in_executor


if TYPE_CHECKING:
    pass


__all__ = ["ObserverOverwatcher"]


class CancelObserveLoopError(GortError):
    """Exception that cancel the observing loop."""

    pass


class ObserverMonitorTask(OverwatcherModuleTask["ObserverOverwatcher"]):
    """Monitors the observer status."""

    name = "observer_monitor"
    keep_alive = False
    restart_on_error = True

    interval: float = 1

    async def task(self):
        """Handles whether we should start the observing loop."""

        # These checks will only start the observing loop. If the weather is
        # unsafe or daytime has been reached the main task will handle stopping
        # the loop.

        state = self.overwatcher.state
        ephemeris = self.overwatcher.ephemeris

        while True:
            if state.dry_run:
                pass

            elif self.module.is_observing or self.module.is_cancelling:
                pass

            elif not state.enabled or state.troubleshooting:
                pass

            elif not ephemeris.ephemeris:
                pass

            elif state.calibrating:
                pass

            elif not state.safe:
                pass

            elif ephemeris.is_night(mode="observer"):
                # Start observing if it's night (after evening twilight) but not too
                # close to the morning twilight.
                try:
                    await self.module.start_observing()
                except Exception as err:
                    await self.notify(
                        "An error occurred while starting the "
                        f"observing loop: {decap(err)}",
                        level="error",
                    )
                    await asyncio.sleep(15)

            await asyncio.sleep(self.interval)


class ObserverOverwatcher(OverwatcherModule):
    name = "observer"
    delay = 5

    tasks = [ObserverMonitorTask()]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.observe_loop: asyncio.Task | None = None
        self.next_exposure_completes: float = 0

        self.focusing: bool = False
        self._starting_observations: bool = False
        self._cancelling: bool = False
        self._schedule_shudtown: bool = False

        self.force_focus: bool = False  # Force focus before the next tile

    @property
    def is_observing(self) -> bool:
        """Returns whether the observer is currently observing."""

        if self._starting_observations:
            return True

        return self.observe_loop is not None and not self.observe_loop.done()

    @property
    def is_cancelling(self) -> bool:
        """Returns whether the observer is currently cancelling."""

        if not self.is_observing:
            self._cancelling = False
            return False

        return self._cancelling

    def cancel(self):
        """Requests the cancellation of the observing loop."""

        if self.is_observing and not self.is_cancelling:
            self._cancelling = True
            self.gort.observer.cancelling = True

    async def start_observing(self):
        """Starts observations."""

        if self.overwatcher.state.dry_run:
            return

        if self.is_observing or self.is_cancelling:
            return

        if self.overwatcher.state.calibrating:
            raise GortError("Cannot start observing while calibrating.")

        if not self.overwatcher.state.safe:
            raise GortError("Cannot safely open the telescope.")

        await self.notify("Starting observations.")

        if not (await self.overwatcher.dome.is_opening()):
            try:
                self._starting_observations = True
                await self.notify("Running the start-up sequence and opening the dome.")
                await self.overwatcher.startup()
            except Exception:
                # If the dome failed to open that will disable the
                # Overwatcher and prevent the startup to run again.
                self.log.error("Startup routine failed.")
                return
            finally:
                self._starting_observations = False

        self.observe_loop = asyncio.create_task(self.observe_loop_task())
        await asyncio.sleep(1)

    async def stop_observing(
        self,
        immediate: bool = False,
        reason: str | None = None,
        block: bool = False,
    ):
        """Stops observations."""

        if self.overwatcher.state.dry_run:
            return

        if not self.is_observing:
            return

        # Prepare the notification message.
        msg = "Stopping observations "
        if immediate:
            msg += "immediately."
        else:
            msg += "after the current tile completes."

        if reason is not None:
            msg += f" Reason: {decap(reason)}"

        await self.overwatcher.notify(ensure_period(msg), level="info")

        # Cancel the observing loop.
        self.cancel()

        # If immediate, cancel the task now and cleanup.
        if immediate:
            if self._starting_observations:
                # Just let the startup finish. observe_loop_task() will immediately
                # return because we have set _is_cancelling=True.
                await self.notify("Waiting for the start-up sequence to finish.")
                if block:
                    while True:
                        if not self._starting_observations:
                            break
                        await asyncio.sleep(1)
                return

            try:
                await cancel_task(self.observe_loop)
            except Exception as err:
                self.log.error(f"Error while cancelling observing loop: {decap(err)}")
            finally:
                self.observe_loop = None

            # The guiders may have been left running or the spectrograph may still
            # be exposing. Clean up to avoid issues.
            await self.gort.cleanup(readout=False)

        # If block, wait for the observing loop to finish.
        if block and self.observe_loop and not self.observe_loop.done():
            await self.observe_loop
            self.observe_loop = None

        # We are not troubleshooting any more if we have stopped the loop, but if we
        # cancelled the task the troubleshooter event may still be cleared.
        self.overwatcher.troubleshooter.reset()

    async def observe_loop_task(self):
        """Runs the observing loop."""

        # Immediately return if we are cancelling. This can happen if we disable
        # the overwatcher during startup.
        if self.is_cancelling:
            return

        # Check that we have not started too early (this happens when we open the dome
        # a bit early to avoid wasting time). If so, wait until we are properly in
        # the observing window.
        if not self.overwatcher.ephemeris.is_night():
            time_to_twilight = self.overwatcher.ephemeris.time_to_evening_twilight()
            if time_to_twilight and time_to_twilight > 0:
                await self.notify(
                    f"Waiting until evening twilight to start observing "
                    f"({time_to_twilight:.0f} seconds)."
                )
                await asyncio.sleep(time_to_twilight)

        await self.gort.cleanup(readout=True)
        observer = self.gort.observer

        n_tile_positions = 0
        self._schedule_shudtown = False

        while True:
            try:
                # Wait in case the troubleshooter is doing something.
                await self.overwatcher.troubleshooter.wait_until_ready(300)

                # We want to avoid re-acquiring the tile between dithers. We call
                # the scheduler here and control the dither position loop ourselves.
                tile: Tile = await run_in_executor(Tile.from_scheduler)

                self.log.info(
                    f"Received tile {tile.tile_id} from scheduler: "
                    f"observing dither positions {tile.dither_positions}."
                )

                if self.is_cancelling:
                    break

                await self.check_focus(force=n_tile_positions == 0 or self.force_focus)

                for dpos in tile.dither_positions:
                    exp: Exposure | bool = False

                    await self.overwatcher.troubleshooter.wait_until_ready(300)

                    if not self.overwatcher.ephemeris.is_night(mode="observer"):
                        await self.notify(
                            "Twilight will be reached before the next exposure "
                            "completes. Stopping observations now."
                        )
                        self.cancel()
                        self._schedule_shudtown = True
                        break

                    # The exposure will complete in 900 seconds + acquisition + readout
                    self.next_exposure_completes = time() + 90 + 900 + 60

                    if not (await self.pre_observe_checks()):
                        raise CancelObserveLoopError("Pre-observe checks failed.")

                    result, exps = await observer.observe_tile(
                        tile=tile,
                        dither_position=dpos,
                        async_readout=True,
                        keep_guiding=True,
                        skip_slew_when_acquired=True,
                        run_cleanup=False,
                        cleanup_on_interrupt=True,
                        show_progress=False,
                    )

                    n_tile_positions += 1

                    # Clear counts of errors that are reset
                    # when a tile is successfully observed.
                    self.overwatcher.troubleshooter.reset()

                    if not result and not self.is_cancelling:
                        raise GortError("The observation ended with error state.")

                    if result and len(exps) > 0:
                        exp = exps[0]

                    try:
                        post_exposure_status = await self.post_exposure(exp)
                    except Exception as err:
                        await self.notify(
                            f"Failed to run post-exposure routine: {decap(err)}",
                            level="error",
                        )
                        post_exposure_status = False

                    if self.is_cancelling or not post_exposure_status:
                        break

            except asyncio.CancelledError:
                break

            except TroubleshooterTimeoutError:
                await self.notify(
                    "The troubleshooter timed out after 300 seconds. "
                    "Cancelling observations.",
                    level="critical",
                )
                break

            except CancelObserveLoopError as err:
                await self.notify(
                    f"Cancelling observations: {decap(err)}",
                    level="error",
                )
                break

            except Exception as err:
                if await self.overwatcher.troubleshooter.handle(err):
                    if self.is_cancelling:
                        break
                    if self.focusing:
                        # Force a new focus if the error occurred while focusing.
                        self.force_focus = True

                    continue

            finally:
                self.exposure_completes = 0

                if self.is_cancelling:
                    try:
                        if exp is not False and not exp.done():
                            self.log.warning("Waiting for exposure to read.")
                            await asyncio.wait_for(exp, timeout=80)
                    except Exception:
                        self.log.error("Failed reading last exposure.")

                    break

        await self.gort.cleanup(readout=False)
        await self.notify("The observing loop has ended.")

        if self._schedule_shudtown:
            # Do not set shutdown_pending until here to prevent the
            # main overwatcher task cancelling the last exposure and
            # various recursion problems.
            self.overwatcher.state.shutdown_pending = True

    async def check_focus(self, force: bool = False):
        """Checks if it's time to focus the telescope."""

        should_focus: bool = False

        # Retry focusing up to 2 times with a 10 second delay and a 150 second timeout.
        focus_retrier = Retrier(max_attempts=2, delay=10, timeout=150)

        if not force:
            # Check if we should refocus.
            focus_info = await self.gort.guiders.sci.get_focus_info()
            focus_age = focus_info["reference_focus"]["age"]

            if focus_age is None or focus_age > 2 * 3600:  # Focus sweeps every 2 hours
                should_focus = True

        else:
            should_focus = True

        # Focus when the loop starts or every 1 hour or at the beginning
        # of the loop.
        if should_focus:
            try:
                self.focusing = True
                await self.notify("Focusing telescopes.")
                await focus_retrier(self.gort.guiders.focus)()
            except Exception as err:
                await self.notify(
                    f"Failed twice while focusing the telescopes: {decap(err)}",
                    level="error",
                )
                await self.notify(
                    "I will continue observations and try to "
                    "focus again after the next tile."
                )
                self.force_focus = True
            else:
                self.force_focus = False
            finally:
                self.focusing = False

    async def pre_observe_checks(self):
        """Runs pre-observe checks."""

        if await self.gort.specs.are_errored():
            exp = self.gort.specs.last_exposure
            if exp and not exp.done():
                self.log.warning(
                    "Spectrographs are idle but an exposure is ongoing. "
                    "Waiting for it to finish before resetting."
                )
                try:
                    await asyncio.wait_for(exp, timeout=100)
                except asyncio.TimeoutError:
                    self.log.error(
                        "Timed out waiting for exposure to finish. "
                        "Resetting spectrographs."
                    )

            await self.gort.specs.reset()

        # Use the code in the troubleshooter to check if all AG cameras are connected.
        acq_failed_recipe = AcquisitionFailedRecipe(self.overwatcher.troubleshooter)

        self.log.debug("Checking AG cameras.")
        ag_pings = await acq_failed_recipe.ping_ag_cameras()
        cameras_alive = await acq_failed_recipe.all_cameras_alive()

        if not all(ag_pings.values()) or not cameras_alive:
            self.log.error("Not all AG cameras ping. Running troubleshooting recipe.")
            try:
                await acq_failed_recipe._handle_disconnected_cameras()
            except TroubleshooterCriticalError as err:
                await self.notify(
                    f"Critical error while checking AG cameras: {decap(err)}",
                    level="error",
                )
                return False

        return True

    async def post_exposure(self, exp: Exposure | bool):
        """Runs post-exposure checks."""

        if self._cancelling:
            return False

        if exp is False:
            raise GortError("No exposure was returned.")

        # Output transparency data for the last exposure.
        transparency = self.overwatcher.transparency
        transparency.write_to_log(["sci"])

        # TODO: for now if the transparency is bad we close and disable the overwatcher.
        # Since we are inside the observer loop we cannot just call shutdown() or
        # we'll get a recursion error, so we cancel the loop and schedule the shutdown.
        if transparency.quality["sci"] & TransparencyQuality.BAD:
            self.cancel()
            await self.notify(
                "Transparency is bad. Stopping observations and closing the "
                "dome. Resume observations manually when the transparency is deemed "
                "good.",
                level="warning",
            )
            self._schedule_shudtown = True

            return False

        return True

        # TODO: monitor transparency and resume observing automatically.
        if transparency.quality["sci"] & TransparencyQuality.BAD:
            await self.notify(
                "Transparency is bad. Stopping observations and starting "
                "the transparency monitor.",
            )

            # If we reach twilight this will cause the overwatcher
            # to immediately stop observations.
            self.exposure_completes = 0

            try:
                await asyncio.wait_for(
                    transparency.start_monitoring(),
                    timeout=3600,
                )

            except asyncio.TimeoutError:
                await self.notify("Transparency monitor timed out.", level="warning")
                await self.overwatcher.shutdown(
                    reason="Transparency has been bad for over one hour.",
                    disable_overwatcher=True,
                )

            else:
                # The transparency monitor has ended. There are two possible reasons:

                # - Something stopped the observing loop and with it the monitor.
                #   Do nothing and return. The main task will handle the rest.
                if self._cancelling:
                    return

                # - The transparency is good and the monitor has ended.
                if transparency.quality["sci"] & TransparencyQuality.GOOD:
                    await self.notify("Transparency is good. Resuming observations.")
                    return

                else:
                    await self.notify(
                        "Transparency is still bad but the monitor stopped. "
                        "Triggering shutdown.",
                    )
                    await self.overwatcher.shutdown(
                        reason="Transparency monitor failed.",
                        disable_overwatcher=True,
                    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-06-15
# @Filename: spec.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING, Sequence

from gort.devices.core import GortDevice, GortDeviceSet
from gort.exceptions import ErrorCode, GortError, GortSpecError
from gort.exposure import Exposure


if TYPE_CHECKING:
    from gort.gort import Gort
    from gort.remote import ActorReply


__all__ = ["Spectrograph", "SpectrographSet", "IEB"]


class IEB(GortDevice):
    """A class representing an Instrument Electronics Box."""

    def __init__(self, gort: Gort, name: str, actor: str):
        super().__init__(gort, name, actor)

        self.spec_name = self.name.split(".")[1]

    async def status(self):
        """Returns the status of the IEB."""

        replies: list[ActorReply] = await asyncio.gather(
            self.actor.commands.shutter.commands.status(),
            self.actor.commands.hartmann.commands.status(),
            self.actor.commands.transducer.commands.status(),
            self.actor.commands.wago.commands.status(),
            self.actor.commands.wago.commands.getpower(),
            return_exceptions=True,
        )

        status = {}
        for reply in replies:
            if isinstance(reply, Exception):
                self.write_to_log(str(reply), "warning")
                continue

            flat_reply = reply.flatten()
            if "transducer" in flat_reply:
                flat_reply = {f"{self.spec_name}_pressures": flat_reply["transducer"]}
            status.update(flat_reply)

        return status

    async def power(self, devices: str | list[str], on: bool = True):
        """Powers on/off the shutter or Hartmann doors.

        Parameters
        ----------
        device
            The device to power on/off. Either ``'shutter'``, ``'hartmann_left'``,
            or ``'hartmann_right'``. Can be a list of devices to modify.
        on
            If :obj:`True` powers on the device; otherwise powers it down.

        """

        if isinstance(devices, str):
            devices = [devices]

        tasks = []
        for device in devices:
            if device in ["hl", "left"]:
                device = "hartmann_left"
            elif device in ["hr", "right"]:
                device = "hartmann_right"

            if device not in ["shutter", "hartmann_left", "hartmann_right"]:
                raise GortSpecError(
                    f"Invalid device {device}.",
                    error_code=ErrorCode.USAGE_ERROR,
                )

            self.write_to_log(f"Powering {'on' if on else 'off'} {device}.", "info")

            tasks.append(
                self.actor.commands.wago.commands.setpower(
                    device,
                    action="ON" if on else "OFF",
                )
            )

        await asyncio.gather(*tasks)

    async def do(self, devices: str | list[str], action: str):
        """Performs an action on a device. Powers the device if needed.

        Parameters
        ----------
        device
            The device to act on. Either ``'shutter'``, ``'hartmann_left'``,
            or ``'hartmann_right'``. Can be a list of devices to modify.
        action
            The action to perform. Can be ``'open'``, ``'close'``, ``'home'``,
            or ``'init'``.

        """

        if isinstance(devices, str):
            devices = [devices]

        if action not in ["open", "close", "home", "init"]:
            raise GortSpecError(
                f"Invalid action {action}.",
                error_code=ErrorCode.USAGE_ERROR,
            )

        status = await self.status()

        tasks = []
        hartmann_done = False
        for device in devices:
            if device in ["hl", "left"]:
                device = "hartmann_left"
            elif device in ["hr", "right"]:
                device = "hartmann_right"

            if device not in ["shutter", "hartmann_left", "hartmann_right"]:
                raise GortSpecError(
                    f"Invalid device {device}.",
                    error_code=ErrorCode.USAGE_ERROR,
                )

            self.write_to_log(f"Performing {action!r} on {device}.", "info")

            if status[f"{self.spec_name}_relays"][device] is False:
                self.write_to_log(f"Device {device} is off. Powering it on.", "warning")
                await self.power(device)

            if "hartmann" in device:
                command = getattr(self.actor.commands.hartmann.commands, action)
                if action in ["home", "init"]:
                    if hartmann_done:
                        # Avoid homing/initialising the HD twice.
                        continue
                    tasks.append(command())
                    hartmann_done = True
                else:
                    tasks.append(command(side=device.split("_")[1]))
            else:
                command = getattr(self.actor.commands.shutter.commands, action)
                tasks.append(command())

        await asyncio.gather(*tasks)

    async def open(self, devices: str | list[str]):
        """Opens a device or list of devices.

        Parameters
        ----------
        device
            The device to act on. Either ``'shutter'``, ``'hartmann_left'``,
            or ``'hartmann_right'``. Can be a list of devices to open.

        """

        await self.do(devices, "open")

    async def close(self, devices: str | list[str]):
        """Closes a device or list of devices.

        Parameters
        ----------
        device
            The device to act on. Either ``'shutter'``, ``'hartmann_left'``,
            or ``'hartmann_right'``. Can be a list of devices to close.

        """

        await self.do(devices, "close")

    async def home(self, devices: str | list[str]):
        """Homes a device or list of devices.

        Parameters
        ----------
        device
            The device to act on. Either ``'shutter'``, ``'hartmann_left'``,
            or ``'hartmann_right'``. Can be a list of devices to home.

        """

        await self.do(devices, "home")

    async def init(self, devices: str | list[str], home: bool = True):
        """Initialises a device or list of devices.

        Parameters
        ----------
        device
            The device to act on. Either ``'shutter'``, ``'hartmann_left'``,
            or ``'hartmann_right'``. Can be a list of devices to initialise.
        home
            If :obj:`True` homes the devices after initialising them.

        """

        await self.do(devices, "init")

        if home:
            await self.do(devices, "home")


class Spectrograph(GortDevice):
    """Class representing an LVM spectrograph functionality."""

    def __init__(self, gort: Gort, name: str, actor: str, **kwargs):
        super().__init__(gort, name, actor)

        self.nps = self.gort.nps[name]
        self.ieb = IEB(gort, f"ieb.{self.name}", f"lvmieb.{self.name}")

    async def status(self, simple: bool = False):
        """Retrieves the status of the telescope.

        Parameters
        ----------
        simple
            If :obj:`True` returns a short version of the status.

        """

        reply: ActorReply = await self.actor.commands.status(simple=simple)
        flatten_reply = reply.flatten()

        return flatten_reply.get("status", {})

    async def is_idle(self):
        """Returns :obj:`True` if the spectrograph is idle and ready to expose."""

        status = await self.status(simple=True)
        names = status["status_names"]

        for status in ["READOUT_PENDING", "ERROR"]:
            if status in names:
                return False

        return "IDLE" in names

    async def is_errored(self):
        """Returns :obj:`True` if the spectrograph is in error."""

        status = await self.status(simple=True)
        return "ERROR" in status["status_names"]

    async def is_exposing(self):
        """Returns :obj:`True` if the spectrograph is exposing."""

        status = await self.status(simple=True)
        return "EXPOSING" in status["status_names"]

    async def is_reading(self):
        """Returns :obj:`True` if the spectrograph is idle and ready to expose."""

        status = await self.status(simple=True)
        return "READING" in status["status_names"]

    async def initialise(self):
        """Initialises the spectrograph and flashes the ACF configuration file."""

        self.write_to_log("Initialising spectrograph and flashing ACF.", "info")
        await self.actor.commands.init()

    async def abort(self):
        """Aborts an ongoing exposure."""

        self.write_to_log("Aborting exposures.", "warning")
        try:
            await self.actor.commands.abort()
        except GortError:
            pass

        await self.actor.commands.reset()

        self.write_to_log("Closing shutter.")
        await self.ieb.close("shutter")

    async def reset(self, full: bool = False):
        """Resets the spectrograph to a valid state."""

        await self.actor.commands.reset()

        mech_status = await self.ieb.status()

        # Always try to close the shutter. If this fails, try a full reset in
        # part to give the shutter another chance but also because it could be
        # that it failed because the power to the motor controllers is off.
        if full is False and mech_status[f"{self.name}_shutter"]["open"]:
            self.write_to_log(f"Closing {self.name} shutter.", "warning")
            try:
                await self.ieb.close("shutter")
            except Exception as ee:
                self.write_to_log(f"Failed closing the shutter: {ee}", "warning")
                self.write_to_log("Trying a full reset.", "warning")
                full = True

        if full:
            relays = mech_status[f"{self.name}_relays"]
            if not all(relays.values()):
                self.write_to_log(f"Powering on {self.name} relays.", "warning")
                await self.ieb.power(list(relays.keys()))
                mech_status = await self.ieb.status()

            if mech_status[f"{self.name}_shutter"]["open"]:
                self.write_to_log(f"Closing {self.name} shutter.", "warning")
                await self.ieb.close("shutter")

            for side in ["left", "right"]:
                if not mech_status[f"{self.name}_hartmann_{side}"]["open"]:
                    self.write_to_log(f"Opening {self.name} {side} HD.", "warning")
                    await self.ieb.open(f"hartmann_{side}")

    async def expose(self, **kwargs):
        """Exposes the spectrograph."""

        if not (await self.is_idle()):
            raise GortSpecError(
                "Spectrographs is not idle. Cannot expose.",
                error_code=ErrorCode.SECTROGRAPH_FAILED_EXPOSING,
            )

        self.write_to_log(f"Exposing spectrograph {self.name}.")

        await self.actor.commands.expose(**kwargs)


class SpectrographSet(GortDeviceSet[Spectrograph]):
    """A set of LVM spectrographs."""

    __DEVICE_CLASS__ = Spectrograph
    __DEPLOYMENTS__ = ["lvmscp"]

    def __init__(self, gort: Gort, data: dict[str, dict], **kwargs):
        super().__init__(gort, data, **kwargs)

        self.last_exposure: Exposure | None = None

    async def status(self, simple: bool = False) -> dict[str, dict]:
        """Collects the status of each spectrograph.

        Parameters
        ----------
        simple
            If :obj:`True` returns a short version of the status.

        """

        names = list(self)
        statuses = await self.call_device_method(Spectrograph.status, simple=simple)

        return dict(zip(names, statuses))

    def get_expno(self):
        """Returns the next exposure sequence number."""

        next_exposure_number_path = self.gort.config["specs"]["nextExposureNumber"]
        with open(next_exposure_number_path, "r") as fd:
            data = fd.read().strip()
            expno = int(data) if data != "" else 1

        return expno

    async def are_idle(self):
        """Returns :obj:`True` if all the spectrographs are idle and ready to expose."""

        return all(await self.call_device_method(Spectrograph.is_idle))

    async def are_reading(self):
        """Returns :obj:`True` if any of the spectrographs are reading."""

        return any(await self.call_device_method(Spectrograph.is_reading))

    async def are_exposing(self):
        """Returns :obj:`True` if any of the spectrographs are exposing."""

        return any(await self.call_device_method(Spectrograph.is_exposing))

    async def are_errored(self):
        """Returns :obj:`True` if any of the spectrographs are errored."""

        return any(await self.call_device_method(Spectrograph.is_errored))

    async def expose(
        self,
        exposure_time: float | None = None,
        flavour: str | None = None,
        header: dict | None = None,
        show_progress: bool | None = None,
        async_readout: bool = False,
        count: int = 1,
        object: str | None = None,
        specs: Sequence[str] | None = None,
        **kwargs,
    ) -> Exposure | list[Exposure]:
        """Exposes the spectrographs.

        Parameters
        ----------
        exposure_time
            The exposure time. If not set, assumes this must
            be a bias.
        flavour
            The exposure type, either ``'object'``, ``'arc'``,
            ``'flat'``, ``'dark'``, or ``'bias'``
        header
            Additional data to add to the headers.
        show_progress
            Displays a progress bar with the elapsed exposure time.
            If :obj:`None` (the default), will show the progress bar only
            in interactive sessions.
        async_readout
            Returns after integration completes. Readout is initiated
            but handled asynchronously and can be await by awaiting
            the returned :obj:`.Exposure` object.
        count
            The number of exposures to take.
        object
            A string that will be stored in the ``OBJECT`` header
            keyword.
        specs
            List the spectrographs to expose. Defaults to all.

        Returns
        -------
        exp_nos
            The numbers of the exposed frames. If ``count`` is greater than
            one, returns a list of exposures.

        """

        if self.last_exposure is not None and not self.last_exposure.done():
            self.write_to_log("Waiting for previous exposure to read out.", "warning")
            await self.last_exposure

        if not (await self.are_idle()):
            raise GortSpecError(
                "Spectrographs are not idle. Cannot expose.",
                error_code=ErrorCode.SECTROGRAPH_NOT_IDLE,
            )

        if count <= 0:
            raise GortSpecError("Invalid count.", error_code=ErrorCode.USAGE_ERROR)

        if exposure_time is None or exposure_time == 0.0:
            flavour = "bias"
            exposure_time = 0.0

        exposures: list[Exposure] = []

        for nexp in range(int(count)):
            exposure = Exposure(self.gort, flavour=flavour, object=object, specs=specs)

            async_readout_this_exp = async_readout if nexp == int(count) - 1 else False

            await exposure.expose(
                exposure_time=exposure_time,
                header=header,
                async_readout=async_readout_this_exp,
                show_progress=show_progress,
                **kwargs,
            )
            exposures.append(exposure)

        if async_readout:
            self.write_to_log("Returning with async readout ongoing.")

        if len(exposures) == 1:
            return exposures[0]

        return exposures

    async def reset(self, full: bool = False):
        """Reset the spectrographs."""

        await self.call_device_method(Spectrograph.reset, full=full)

        self.last_exposure = None

    async def initialise(self):
        """Initialises the spectrographs and flashes the ACF configuration file."""

        await self.call_device_method(Spectrograph.initialise)

    async def abort(self):
        """Aborts an ongoing exposure."""

        await self.call_device_method(Spectrograph.abort)
        self.last_exposure = None
